from typing import Dict, Generator, Tuple, Optional, Union

import pandas as pd
import torch
import torchtext

from .torch_data import toTensor, TorchDataSet, TorchDataSetProvider


class TorchtextDataSetFromDataFrame(torchtext.data.Dataset):
    """
    A specialisation of torchtext.data.Dataset, where the data is taken from a pandas.DataFrame
    """
    def __init__(self, df: pd.DataFrame, fields: Dict[str, torchtext.data.Field]):
        """
        :param df: the data frame from which to obtain the data
        :param fields: a mapping from column names in the given data frame to torchtext fields, i.e.
            the keys are the columns to read and the values are the fields to use for generated Example instances
        """
        examples = df.apply(self._exampleFromSeries, args=(fields,), axis=1).tolist()
        fields = dict(fields)
        super().__init__(examples, fields)

    @classmethod
    def _exampleFromSeries(cls, series: pd.Series, fields: Dict[str, torchtext.data.Field]):
        return cls._exampleFromDict(series.to_dict(), fields)

    @classmethod
    def _exampleFromDict(cls, d: dict, fields: Dict[str, torchtext.data.Field]):
        ex = torchtext.data.Example()
        for key, field in fields.items():
            if key not in d:
                raise ValueError("Specified key {} was not found in "
                                 "the input data".format(key))
            if field is not None:
                setattr(ex, key, field.preprocess(d[key]))
            else:
                setattr(ex, key, d[key])
        return ex


class TorchDataSetFromTorchtextDataSet(TorchDataSet):
    def __init__(self, dataSet: torchtext.data.Dataset, inputField: str, outputField: Optional[str], cuda: bool):
        self.outputField = outputField
        self.inputField = inputField
        self.dataSet = dataSet
        self.cuda = cuda

    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False) -> Generator[Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor], None, None]:
        iterator = torchtext.data.BucketIterator(self.dataSet,
            batch_size=batchSize,
            sort_key=lambda x: len(x.text),
            sort_within_batch=False)

        for batch in iterator:
            x = toTensor(getattr(batch, self.inputField), self.cuda)
            if not inputOnly and self.outputField is not None:
                y = toTensor(getattr(batch, self.outputField), self.cuda)
                yield x, y
            else:
                yield x

    def size(self) -> Optional[int]:
        return len(self.dataSet)


class TorchDataSetProviderFromTorchtextDataSet(TorchDataSetProvider):
    def __init__(self, dataSet: torchtext.data.Dataset, inputField: str, outputField: str, cuda: bool, modelOutputDim, inputDim=None):
        super().__init__(modelOutputDim=modelOutputDim, inputDim=inputDim)
        self.dataSet = dataSet
        self.outputField = outputField
        self.inputField = inputField
        self.cuda = cuda

    def provideSplit(self, fractionalSizeOfFirstSet: float) -> Tuple[TorchDataSet, TorchDataSet]:
        d1, d2 = self.dataSet.split(fractionalSizeOfFirstSet)
        return self._createDataSet(d1), self._createDataSet(d2)

    def _createDataSet(self, d: torchtext.data.Dataset):
        return TorchDataSetFromTorchtextDataSet(d, self.inputField, self.outputField, self.cuda)


from abc import ABC, abstractmethod
from typing import Tuple, Sequence

import numpy as np
import pandas as pd
import scipy.stats


class InputOutputData:
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        if len(inputs) != len(outputs):
            raise ValueError("Lengths do not match")
        self.inputs = inputs
        self.outputs = outputs

    def __len__(self):
        return len(self.inputs)

    @property
    def inputDim(self):
        return self.inputs.shape[1]

    @property
    def outputDim(self):
        return self.outputs.shape[1]

    def filterIndices(self, indices: Sequence[int]) -> 'InputOutputData':
        inputs = self.inputs.iloc[indices]
        outputs = self.outputs.iloc[indices]
        return InputOutputData(inputs, outputs)

    def computeInputOutputCorrelation(self):
        correlations = {}
        for outputCol in self.outputs.columns:
            correlations[outputCol] = {}
            outputSeries = self.outputs[outputCol]
            for inputCol in self.inputs.columns:
                inputSeries = self.inputs[inputCol]
                pcc, pvalue = scipy.stats.pearsonr(inputSeries, outputSeries)
                correlations[outputCol][inputCol] = pcc
        return correlations


class DataSplitter(ABC):
    @abstractmethod
    def split(self, data: InputOutputData) -> Tuple[InputOutputData, InputOutputData]:
        pass


class DataSplitterFractional(DataSplitter):
    def __init__(self, fractionalSizeOfFirstSet, shuffle=True, randomSeed: int = 42):
        if not 0 <= fractionalSizeOfFirstSet <= 1:
            raise Exception(f"invalid fraction: {fractionalSizeOfFirstSet}")
        self.fractionalSizeOfFirstSet = fractionalSizeOfFirstSet
        self.shuffle = shuffle
        self.randomSeed = randomSeed

    def split(self, data: InputOutputData) -> Tuple[InputOutputData, InputOutputData]:
        numDataPoints = len(data)
        splitIndex = int(numDataPoints * self.fractionalSizeOfFirstSet)
        rand = np.random.RandomState(self.randomSeed)
        if self.shuffle:
            indices = rand.permutation(numDataPoints)
        else:
            indices = range(numDataPoints)
        indicesA = indices[:splitIndex]
        indicesB = indices[splitIndex:]
        A = data.filterIndices(list(indicesA))
        B = data.filterIndices(list(indicesB))
        return A, B



import logging
import time
from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Generator, Generic, TypeVar, Sequence

import pandas as pd

from .eval_stats.eval_stats_base import EvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats, ClassificationMetric
from .eval_stats.eval_stats_regression import RegressionEvalStats, RegressionEvalStatsCollection, RegressionMetric
from ..data_ingest import DataSplitter, DataSplitterFractional, InputOutputData
from ..util.typing import PandasNamedTuple
from ..vector_model import VectorClassificationModel, VectorModel, PredictorModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=EvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)


class MetricsDictProvider(ABC):
    @abstractmethod
    def computeMetrics(self, model, **kwargs) -> Dict[str, float]:
        pass


class PredictorModelEvaluationData(ABC, Generic[TEvalStats]):
    def __init__(self, statsDict: Dict[str, TEvalStats], inputData: pd.DataFrame, model: PredictorModel):
        """
        :param statsDict: a dictionary mapping from output variable name to the evaluation statistics object
        :param inputData: the input data that was used to produce the results
        :param model: the model that was used to produce predictions
        """
        self.inputData = inputData
        self.evalStatsByVarName = statsDict
        self.predictedVarNames = list(self.evalStatsByVarName.keys())
        self.modelName = model.getName()

    def getEvalStats(self, predictedVarName=None) -> TEvalStats:
        if predictedVarName is None:
            if len(self.evalStatsByVarName) != 1:
                raise Exception(f"Must provide name of predicted variable name, as multiple variables were predicted {list(self.evalStatsByVarName.keys())}")
            else:
                predictedVarName = next(iter(self.evalStatsByVarName.keys()))
        evalStats = self.evalStatsByVarName.get(predictedVarName)
        if evalStats is None:
            raise ValueError(f"No evaluation data present for '{predictedVarName}'; known output variables: {list(self.evalStatsByVarName.keys())}")
        return evalStats

    def getDataFrame(self):
        """
        Returns an DataFrame with all evaluation metrics (one row per output variable)

        :return: a DataFrame containing evaluation metrics
        """
        statsDicts = []
        varNames = []
        for predictedVarName, evalStats in self.evalStatsByVarName.items():
            statsDicts.append(evalStats.getAll())
            varNames.append(predictedVarName)
        df = pd.DataFrame(statsDicts, index=varNames)
        df.index.name = "predictedVar"
        return df

    def iterInputOutputGroundTruthTuples(self, predictedVarName=None) -> Generator[Tuple[PandasNamedTuple, Any, Any], None, None]:
        evalStats = self.getEvalStats(predictedVarName)
        for i, namedTuple in enumerate(self.inputData.itertuples()):
            yield namedTuple, evalStats.y_predicted[i], evalStats.y_true[i]


class VectorRegressionModelEvaluationData(PredictorModelEvaluationData[RegressionEvalStats]):
    def getEvalStatsCollection(self):
        return RegressionEvalStatsCollection(list(self.evalStatsByVarName.values()))


class VectorModelEvaluator(MetricsDictProvider, ABC):
    def __init__(self, data: InputOutputData, testData: InputOutputData = None, dataSplitter: DataSplitter = None,
            testFraction=None, randomSeed=42, shuffle=True):
        """
        Constructs an evaluator with test and training data.
        Exactly one of the parameters {testFraction, testData, } must be given

        :param data: the full data set, or, if testData is given, the training data
        :param testData: the data to use for testing/evaluation; if None, must specify either dataSplitter testFraction or dataSplitter
        :param dataSplitter: [if testData is None] a splitter to use in order to obtain; if None, must specify either testData or testFraction
        :param testFraction: [if testData is None, dataSplitter is None] the fraction of the data to use for testing/evaluation;
            if None, must specify either testData or dataSplitter
        :param randomSeed: [if data is None, dataSplitter is None] the random seed to use for the fractional split of the data
        :param shuffle: [if data is None, dataSplitter is None] whether to randomly (based on randomSeed) shuffle the dataset before
            splitting it
        """
        if (testData, dataSplitter, testFraction).count(None) != 2:
            raise ValueError("Exactly one of {testData, dataSplitter, testFraction} must be given")
        if testData is None:
            if dataSplitter is None:
                dataSplitter = DataSplitterFractional(1 - testFraction, shuffle=shuffle, randomSeed=randomSeed)
            self.trainingData, self.testData = dataSplitter.split(data)
        else:
            self.trainingData = data
            self.testData = testData

    def fitModel(self, model: VectorModel):
        """Fits the given model's parameters using this evaluator's training data"""
        startTime = time.time()
        model.fit(self.trainingData.inputs, self.trainingData.outputs)
        log.info(f"Training of {model.__class__.__name__} completed in {time.time() - startTime:.1f} seconds")

    @abstractmethod
    def evalModel(self, model: PredictorModel, onTrainingData=False) -> PredictorModelEvaluationData:
        """
        Evaluates the given model

        :param model: the model to evaluate
        :param onTrainingData: if True, evaluate on this evaluator's training data rather than the held-out test data
        :return: the evaluation result
        """
        pass

    def computeMetrics(self, model: PredictorModel, onTrainingData=False) -> Dict[str, float]:
        evalData = self.evalModel(model)
        return evalData.getEvalStats().getAll()


class VectorRegressionModelEvaluator(VectorModelEvaluator):
    def __init__(self, data: InputOutputData, testData: InputOutputData = None, dataSplitter=None, testFraction=None, randomSeed=42, shuffle=True,
            additionalMetrics: Sequence[RegressionMetric] = None):
        super().__init__(data=data, dataSplitter=dataSplitter, testFraction=testFraction, testData=testData, randomSeed=randomSeed, shuffle=shuffle)
        self.additionalMetrics = additionalMetrics

    def evalModel(self, model: PredictorModel, onTrainingData=False) -> VectorRegressionModelEvaluationData:
        if not model.isRegressionModel():
            raise ValueError(f"Expected a regression model, got {model}")
        evalStatsByVarName = {}
        inputOutputData = self.trainingData if onTrainingData else self.testData
        predictions, groundTruth = self._computeOutputs(model, inputOutputData)
        for predictedVarName in model.getPredictedVariableNames():
            evalStats = RegressionEvalStats(y_predicted=predictions[predictedVarName], y_true=groundTruth[predictedVarName],
                additionalMetrics=self.additionalMetrics)
            evalStatsByVarName[predictedVarName] = evalStats
        return VectorRegressionModelEvaluationData(evalStatsByVarName, inputOutputData.inputs, model)

    def computeTestDataOutputs(self, model: PredictorModel) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the test data

        :param model: the model to apply
        :return: a pair (predictions, groundTruth)
        """
        return self._computeOutputs(model, self.testData)

    def _computeOutputs(self, model, inputOutputData: InputOutputData):
        """
        Applies the given model to the given data

        :param model: the model to apply
        :param inputOutputData: the data set
        :return: a pair (predictions, groundTruth)
        """
        predictions = model.predict(inputOutputData.inputs)
        groundTruth = inputOutputData.outputs
        return predictions, groundTruth


class VectorClassificationModelEvaluationData(PredictorModelEvaluationData[ClassificationEvalStats]):
    pass


class VectorClassificationModelEvaluator(VectorModelEvaluator):
    def __init__(self, data: InputOutputData, testData: InputOutputData = None, dataSplitter=None, testFraction=None,
            randomSeed=42, computeProbabilities=False, shuffle=True, additionalMetrics: Sequence[ClassificationMetric] = None):
        super().__init__(data=data, testData=testData, dataSplitter=dataSplitter, testFraction=testFraction, randomSeed=randomSeed, shuffle=shuffle)
        self.computeProbabilities = computeProbabilities
        self.additionalMetrics = additionalMetrics

    def evalModel(self, model: VectorClassificationModel, onTrainingData=False) -> VectorClassificationModelEvaluationData:
        if model.isRegressionModel():
            raise ValueError(f"Expected a classification model, got {model}")
        inputOutputData = self.trainingData if onTrainingData else self.testData
        predictions, predictions_proba, groundTruth = self._computeOutputs(model, inputOutputData)
        evalStats = ClassificationEvalStats(y_predictedClassProbabilities=predictions_proba, y_predicted=predictions, y_true=groundTruth,
            labels=model.getClassLabels(), additionalMetrics=self.additionalMetrics)
        predictedVarName = model.getPredictedVariableNames()[0]
        return VectorClassificationModelEvaluationData({predictedVarName: evalStats}, inputOutputData.inputs, model)

    def computeTestDataOutputs(self, model: VectorClassificationModel) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the test data

        :param model: the model to apply
        :return: a triple (predictions, predicted class probability vectors, groundTruth) of DataFrames
        """
        return self._computeOutputs(model, self.testData)

    def _computeOutputs(self, model, inputOutputData: InputOutputData) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the given data

        :param model: the model to apply
        :param inputOutputData: the data set
        :return: a triple (predictions, predicted class probability vectors, groundTruth) of DataFrames
        """
        if self.computeProbabilities:
            classProbabilities = model.predictClassProbabilities(inputOutputData.inputs)
            if classProbabilities is None:
                raise Exception(f"Requested computation of class probabilities for a model which does not support it: {model} returned None")
            predictions = model.convertClassProbabilitiesToPredictions(classProbabilities)
        else:
            classProbabilities = None
            predictions = model.predict(inputOutputData.inputs)
        groundTruth = inputOutputData.outputs
        return predictions, classProbabilities, groundTruth



import collections
import re
import numpy as np
import pandas as pd
from typing import Optional
import logging

import torch

from ..util.string import objectRepr
from . import torch_modules
from .torch_base import VectorTorchModel, TorchVectorRegressionModel, TorchVectorClassificationModel
from .torch_data import TorchDataSetProviderFromDataUtil, DataUtil, TensorScaler, TensorScalerIdentity
from ..normalisation import NormalisationMode


log = logging.getLogger(__name__)


class MultiLayerPerceptronTorchModel(VectorTorchModel):
    def __init__(self, cuda, hiddenDims, hidActivationFunction, outputActivationFunction, pDropout=None):
        super().__init__(cuda=cuda)
        self.hidActivationFunction = hidActivationFunction
        self.outputActivationFunction = outputActivationFunction
        self.hiddenDims = hiddenDims
        self.pDropout = pDropout

    def __str__(self):
        return f"_MLP[hiddenDims={self.hiddenDims}, hidAct={self.hidActivationFunction.__name__}, outAct={self.outputActivationFunction.__name__ if self.outputActivationFunction is not None else None}, pDropout={self.pDropout}]"

    def createTorchModuleForDims(self, inputDim, outputDim):
        return torch_modules.MultiLayerPerceptron(inputDim, outputDim, self.hiddenDims,
            hidActivationFn=self.hidActivationFunction, outputActivationFn=self.outputActivationFunction,
            pDropout=self.pDropout)


class MultiLayerPerceptronVectorRegressionModel(TorchVectorRegressionModel):
    def __init__(self, hiddenDims=(5, 5), hidActivationFunction=torch.sigmoid, outputActivationFunction=None,
            normalisationMode=NormalisationMode.MAX_BY_COLUMN,
            cuda=True, pDropout=None, **nnOptimiserParams):
        """
        :param hiddenDims: sequence containing the number of neurons to use in hidden layers
        :param hidActivationFunction: the activation function (torch.*) to use for all hidden layers
        :param outputActivationFunction: the output activation function (torch.* or None)
        :param normalisationMode: the normalisation mode to apply to input and output data
        :param cuda: whether to use CUDA (GPU acceleration)
        :param pDropout: the probability with which to apply dropouts after each hidden layer
        :param nnOptimiserParams: parameters to pass on to NNOptimiser
        """
        super().__init__(MultiLayerPerceptronTorchModel, [cuda, hiddenDims, hidActivationFunction, outputActivationFunction],
                dict(pDropout=pDropout), normalisationMode, nnOptimiserParams)


class MultiLayerPerceptronVectorClassificationModel(TorchVectorClassificationModel):
    def __init__(self, hiddenDims=(5, 5), hidActivationFunction=torch.sigmoid, outputActivationFunction=torch.sigmoid,
            normalisationMode=NormalisationMode.MAX_BY_COLUMN, cuda=True, pDropout=None,
            **nnOptimiserParams):
        """
        :param hiddenDims: sequence containing the number of neurons to use in hidden layers
        :param hidActivationFunction: the activation function (torch.*) to use for all hidden layers
        :param outputActivationFunction: the output activation function (torch.*)
        :param normalisationMode: the normalisation mode to apply to input and output data
        :param cuda: whether to use CUDA (GPU acceleration)
        :param pDropout: the probability with which to apply dropouts after each hidden layer
        :param nnOptimiserParams: parameters to pass on to NNOptimiser
        """
        super().__init__(MultiLayerPerceptronTorchModel, [cuda, hiddenDims, hidActivationFunction, outputActivationFunction],
            dict(pDropout=pDropout), normalisationMode, nnOptimiserParams)


class LSTNetworkVectorClassificationModel(TorchVectorClassificationModel):
    """
    Classification model for time series data using the LSTNetwork architecture.

    Since the model takes a time series as input, it requires that input data frames to use special naming of columns
    such that the data can be interpreted correctly:
    Each column name must start with an N-digit prefix indicating the time slice the data pertains to (for any fixed N);
    the following suffix shall indicate the name of the actual feature.
    For each N-digit prefix, we must have the same set of suffixes in the list of columns, i.e. we must have the same
    features for each time slice in the input time series.
    """
    def __init__(self, numInputTimeSlices, inputDimPerTimeSlice, numClasses: Optional[int] = None,
            numConvolutions: int = 100, numCnnTimeSlices: int = 6, hidRNN: int = 100, skip: int = 0, hidSkip: int = 5,
            hwWindow: int = 0, hwCombine: str = "plus", dropout=0.2, outputActivation="sigmoid",
            nnOptimiserParams: dict = None):
        """
        :param numInputTimeSlices: the number of input time slices
        :param inputDimPerTimeSlice: the dimension of the input data per time slice
        :param numClasses: the number of classes considered by this classification problem; if None, determine from data
        :param numCnnTimeSlices: the number of time slices considered by each convolution (i.e. it is one of the dimensions of the matrix used for
            convolutions, the other dimension being inputDimPerTimeSlice), a.k.a. "Ck"
        :param numConvolutions: the number of separate convolutions to apply, i.e. the number of independent convolution matrices, a.k.a "hidC";
            if it is 0, then the entire complex processing path is not applied.
        :param hidRNN: the number of hidden output dimensions for the RNN stage
        :param skip: the number of time slices to skip for the skip-RNN. If it is 0, then the skip-RNN is not used.
        :param hidSkip: the number of output dimensions of each of the skip parallel RNNs
        :param hwWindow: the number of time slices from the end of the input time series to consider as input for the highway component.
            If it is 0, the highway component is not used.
        :param hwCombine: {"plus", "product", "bilinear"} the function with which the highway component's output is combined with the complex path's output
        :param dropout: the dropout probability to use during training (dropouts are applied after every major step in the evaluation path)
        :param outputActivation: the output activation function
        :param nnOptimiserParams: parameters of NNOptimiser to use for training
        """
        if nnOptimiserParams is None:
            nnOptimiserParams = {}
        self.cuda = nnOptimiserParams.get("cuda", False)
        self.numClasses = numClasses
        lstnetArgs = dict(numInputTimeSlices=numInputTimeSlices, inputDimPerTimeSlice=inputDimPerTimeSlice, numOutputTimeSlices=1,
            outputDimPerTimeSlice=numClasses, numConvolutions=numConvolutions, numCnnTimeSlices=numCnnTimeSlices, hidRNN=hidRNN,
            hwWindow=hwWindow, hwCombine=hwCombine, dropout=dropout, outputActivation=outputActivation,
            skip=skip, hidSkip=hidSkip, isClassification=True)
        super().__init__(self._LSTNetworkModel, modelArgs=[self.cuda], modelKwArgs=lstnetArgs, nnOptimiserParams=nnOptimiserParams)

    class _LSTNetworkModel(VectorTorchModel):
        def __init__(self, cuda, **lstnetArgs):
            """
            :param cuda: flag indicating whether cuda is used
            :param inputDim: the total number of inputs per data point
            :param numClasses: the number of classes to predict
            :param lstnetArgs: arguments with which to construct the underlying LSTNetwork instance
            """
            super().__init__(cuda)
            self.lstnetArgs = lstnetArgs

        def createTorchModuleForDims(self, inputDim, outputDim):
            expectedInputDim = self.lstnetArgs["numInputTimeSlices"] * self.lstnetArgs["inputDimPerTimeSlice"]
            if expectedInputDim != inputDim:
                raise ValueError(f"Unexpected input size {inputDim}, expected {self.inputDim}")
            if self.lstnetArgs["outputDimPerTimeSlice"] is None:
                self.lstnetArgs["outputDimPerTimeSlice"] = outputDim
            else:
                if self.lstnetArgs["outputDimPerTimeSlice"] != outputDim:
                    raise ValueError(f"Unexpected output size {outputDim}, expected {self.lstnetArgs['outputDimPerTimeSlice']}")
            return torch_modules.LSTNetwork(**self.lstnetArgs)

        def __str__(self):
            return objectRepr(self, self.lstnetArgs)

    def _createDataSetProvider(self, inputs: pd.DataFrame, outputs: pd.DataFrame) -> TorchDataSetProviderFromDataUtil:
        if self.numClasses is None:
            self.numClasses = len(self._labels)
        elif self.numClasses != len(self._labels):
            raise ValueError(f"Output dimension {self.numClasses} per time time slice was specified, while the training data contains {len(self._labels)} classes")
        return TorchDataSetProviderFromDataUtil(self.DataUtil(inputs, outputs, self.numClasses), self.cuda)

    def _predictOutputsForInputDataFrame(self, inputs: pd.DataFrame) -> np.ndarray:
        log.info(f"Predicting outputs for {len(inputs)} inputs")
        result = super()._predictOutputsForInputDataFrame(inputs)
        return np.squeeze(result, 2)

    def _computeInputs(self, x: pd.DataFrame, y=None) -> pd.DataFrame:
        x = super()._computeInputs(x, y=y)

        # sort input data frame columns by name
        x = x[sorted(x.columns)]

        # check input column name format and consistency
        colNameRegex = re.compile(r"(\d+).+")
        colsByTimeSlice = collections.defaultdict(list)
        numDigits = None
        for colName in x.columns:
            match = colNameRegex.fullmatch(colName)
            if not match:
                raise ValueError(f"Column name '{colName}' does not match the required format (N-digit prefix indicating the time slice order followed by feature name; for any fixed N); columns={list(x.columns)}")
            timeSlice = match.group(1)
            if numDigits is None:
                numDigits = len(timeSlice)
            elif numDigits != len(timeSlice):
                raise ValueError(f"Inconsistent number of digits in column names: Got {numDigits} leading digits for one feature and {len(timeSlice)} for another; columns={list(x.columns)}")
            colsByTimeSlice[timeSlice].append(colName[numDigits:])
        referenceCols = None
        for timeSlice, cols in colsByTimeSlice.items():
            if referenceCols is None:
                referenceCols = cols
            elif referenceCols != cols:
                raise ValueError(f"Inconsistent features across time slices: Got suffixes {cols} for one time slice and {referenceCols} for another; columns={list(x.columns)}")

        return x

    class DataUtil(DataUtil):
        def __init__(self, x_data: pd.DataFrame, y_data: pd.DataFrame, numClasses):
            self.y_data = y_data
            self.x_data = x_data
            self.numClasses = numClasses
            self.scaler = TensorScalerIdentity()

        def inputDim(self):
            return len(self.x_data.columns)

        def modelOutputDim(self) -> int:
            return self.numClasses

        def splitInputOutputPairs(self, fractionalSizeOfFirstSet):
            splitIndex = round(fractionalSizeOfFirstSet * len(self.y_data))
            y1, x1 = self.getInputOutputPair(self.y_data[:splitIndex], self.x_data[:splitIndex])
            y2, x2 = self.getInputOutputPair(self.y_data[splitIndex:], self.x_data[splitIndex:])
            return (x1, y1), (x2, y2)

        def getInputOutputPair(self, output, input):
            y = torch.tensor(output.values).long()
            x = torch.tensor(input.values).float()
            return y, x

        def getOutputTensorScaler(self) -> TensorScaler:
            return self.scaler

        def getInputTensorScaler(self) -> TensorScaler:
            return self.scaler



from abc import ABC, abstractmethod
import logging
from typing import Any, Union, Optional

import numpy as np
import pandas as pd

from .util.cache import PersistentKeyValueCache


_log = logging.getLogger(__name__)


class ColumnGenerator:
    """
    Generates a single column (pd.Series) from an input data frame, which is to have the same index as the input
    """
    def __init__(self, generatedColumnName: str):
        """
        :param generatedColumnName: the name of the column being generated
        """
        self.generatedColumnName = generatedColumnName

    def generateColumn(self, df: pd.DataFrame) -> pd.Series:
        """
        Generates a column from the input data frame

        :param df: the input data frame
        :return: the column as a named series, which has the same index as the input
        """
        result = self._generateColumn(df)
        if isinstance(result, pd.Series):
            result.name = self.generatedColumnName
        else:
            result = pd.Series(result, index=df.index, name=self.generatedColumnName)
        return result

    @abstractmethod
    def _generateColumn(self, df: pd.DataFrame) -> Union[pd.Series, list, np.ndarray]:
        """
        Performs the actual column generation

        :param df: the input data frame
        :return: a list/array of the same length as df or a series with the same index
        """
        pass


class IndexCachedColumnGenerator(ColumnGenerator):
    """
    Decorator for a column generator which adds support for cached column generation where cache keys are given by the input data frame's index.
    Entries not found in the cache are computed by the wrapped column generator.

    The main use case for this class is to add caching to existing ColumnGenerators. For creating a new caching
    ColumnGenerator the use of ColumnGeneratorCachedByIndex is encouraged.
    """

    _log = _log.getChild(__qualname__)

    def __init__(self, columnGenerator: ColumnGenerator, cache: PersistentKeyValueCache):
        """
        :param columnGenerator: the column generator with which to generate values for keys not found in the cache
        :param cache: the cache in which to store key-value pairs
        """
        super().__init__(columnGenerator.generatedColumnName)
        self.columnGenerator = columnGenerator
        self.cache = cache

    def _generateColumn(self, df: pd.DataFrame) -> pd.Series:
        # compute series of cached values
        cacheValues = [self.cache.get(nt.Index) for nt in df.itertuples()]
        cacheSeries = pd.Series(cacheValues, dtype=object, index=df.index).dropna()

        # compute missing values (if any) via wrapped generator, storing them in the cache
        missingValuesDF = df[~df.index.isin(cacheSeries.index)]
        self._log.info(f"Retrieved {len(cacheSeries)} values from the cache, {len(missingValuesDF)} still to be computed by {self.columnGenerator}")
        if len(missingValuesDF) == 0:
            return cacheSeries
        else:
            missingSeries = self.columnGenerator.generateColumn(missingValuesDF)
            for key, value in missingSeries.iteritems():
                self.cache.set(key, value)
            return pd.concat((cacheSeries, missingSeries))


class ColumnGeneratorCachedByIndex(ColumnGenerator, ABC):
    """
    Base class for column generators, which supports cached column generation, each value being generated independently.
    Cache keys are given by the input data frame's index.
    """

    _log = _log.getChild(__qualname__)

    def __init__(self, generatedColumnName: str, cache: Optional[PersistentKeyValueCache], persistCache=False):
        """
        :param generatedColumnName: the name of the column being generated
        :param cache: the cache in which to store key-value pairs. If None, caching will be disabled
        :param persistCache: whether to persist the cache when pickling
        """
        super().__init__(generatedColumnName)
        self.cache = cache
        self.persistCache = persistCache

    def _generateColumn(self, df: pd.DataFrame) -> Union[pd.Series, list, np.ndarray]:
        self._log.info(f"Generating column {self.generatedColumnName} with {self.__class__.__name__}")
        values = []
        cacheHits = 0
        columnLength = len(df)
        percentageToLog = 0
        for i, namedTuple in enumerate(df.itertuples()):
            percentageGenerated = int(100*i/columnLength)
            if percentageGenerated == percentageToLog:
                self._log.debug(f"Processed {percentageToLog}% of {self.generatedColumnName}")
                percentageToLog += 5

            key = namedTuple.Index
            if self.cache is not None:
                value = self.cache.get(key)
                if value is None:
                    value = self._generateValue(namedTuple)
                    self.cache.set(key, value)
                else:
                    cacheHits += 1
            else:
                value = self._generateValue(namedTuple)
            values.append(value)
        if self.cache is not None:
            self._log.info(f"Cached column generation resulted in {cacheHits}/{columnLength} cache hits")
        return values

    def __getstate__(self):
        if not self.persistCache:
            d = self.__dict__.copy()
            d["cache"] = None
            return d
        return self.__dict__

    @abstractmethod
    def _generateValue(self, namedTuple) -> Any:
        pass



import logging
import time
from typing import Sequence, TypeVar, List

from . import cache

T = TypeVar("T")
log = logging.getLogger(__name__)


def concatSequences(seqs: Sequence[Sequence[T]]) -> List[T]:
    result = []
    for s in seqs:
        result.extend(s)
    return result


def dict2OrderedTuples(d: dict):
    keys = sorted(d.keys())
    values = [d[k] for k in keys]
    return tuple(keys), tuple(values)


class LogTime:
    def __init__(self, name):
        self.name = name
        self.startTime = None

    def start(self):
        self.startTime = time.time()

    def stop(self):
        log.info(f"{self.name} completed in {time.time()-self.startTime} seconds")

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __enter__(self):
        return self.start()



import copy
import logging
import warnings
from abc import ABC, abstractmethod
from typing import Tuple, Any, Generator, Generic, TypeVar, List

import numpy as np

from .eval_stats.eval_stats_base import PredictionEvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats, ClassificationEvalStatsCollection
from .eval_stats.eval_stats_regression import RegressionEvalStats, RegressionEvalStatsCollection
from .evaluator import VectorRegressionModelEvaluationData, VectorClassificationModelEvaluationData, \
    PredictorModelEvaluationData, VectorClassificationModelEvaluator, VectorRegressionModelEvaluator, \
    MetricsDictProvider
from ..data_ingest import InputOutputData
from ..util.typing import PandasNamedTuple
from ..vector_model import VectorClassificationModel, VectorRegressionModel, VectorModel, PredictorModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=PredictorModel)
TEvalStats = TypeVar("TEvalStats", bound=PredictionEvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)
TEvalData = TypeVar("TEvalData", bound=PredictorModelEvaluationData)


class PredictorModelCrossValidationData(ABC, Generic[TModel, TEvalData, TEvalStats, TEvalStatsCollection]):
    def __init__(self, predictorModels: List[TModel], evalDataList: List[TEvalData], predictedVarNames: List[str], testIndicesList=None):
        self.predictedVarNames = predictedVarNames
        self.predictorModels = predictorModels
        self.evalDataList = evalDataList
        self.testIndicesList = testIndicesList

    @property
    def trainedModels(self):
        warnings.warn("Use predictorModels instead, the field trainedModels will be removed in a future release",
                      DeprecationWarning)
        return self.predictorModels

    @property
    def modelName(self):
        return self.evalDataList[0].modelName

    @abstractmethod
    def _createEvalStatsCollection(self, l: List[TEvalStats]) -> TEvalStatsCollection:
        pass

    def getEvalStatsCollection(self, predictedVarName=None) -> TEvalStatsCollection:
        if predictedVarName is None:
            if len(self.predictedVarNames) != 1:
                raise Exception("Must provide name of predicted variable")
            else:
                predictedVarName = self.predictedVarNames[0]
        evalStatsList = [evalData.getEvalStats(predictedVarName) for evalData in self.evalDataList]
        return self._createEvalStatsCollection(evalStatsList)

    def iterInputOutputGroundTruthTuples(self, predictedVarName=None) -> Generator[Tuple[PandasNamedTuple, Any, Any], None, None]:
        for evalData in self.evalDataList:
            evalStats = evalData.getEvalStats(predictedVarName)
            for i, namedTuple in enumerate(evalData.inputData.itertuples()):
                yield namedTuple, evalStats.y_predicted[i], evalStats.y_true[i]


TCrossValData = TypeVar("TCrossValData", bound=PredictorModelCrossValidationData)


class VectorModelCrossValidator(MetricsDictProvider, Generic[TCrossValData], ABC):
    def __init__(self, data: InputOutputData, folds: int = 5, randomSeed=42, returnTrainedModels=False, evaluatorParams: dict = None):
        """
        :param data: the data set
        :param folds: the number of folds
        :param randomSeed: the random seed to use
        :param returnTrainedModels: whether to create a copy of the model for each fold and return each of the models
            (requires that models can be deep-copied); if False, the model that is passed to evalModel is fitted several times
        :param evaluatorParams: keyword parameters with which to instantiate model evaluators
        """
        self.returnTrainedModels = returnTrainedModels
        self.evaluatorParams = evaluatorParams if evaluatorParams is not None else {}
        numDataPoints = len(data)
        permutedIndices = np.random.RandomState(randomSeed).permutation(numDataPoints)
        numTestPoints = numDataPoints // folds
        self.modelEvaluators = []
        for i in range(folds):
            testStartIdx = i * numTestPoints
            testEndIdx = testStartIdx + numTestPoints
            testIndices = permutedIndices[testStartIdx:testEndIdx]
            trainIndices = np.concatenate((permutedIndices[:testStartIdx], permutedIndices[testEndIdx:]))
            self.modelEvaluators.append(self._createModelEvaluator(data.filterIndices(trainIndices), data.filterIndices(testIndices)))

    @abstractmethod
    def _createModelEvaluator(self, trainingData: InputOutputData, testData: InputOutputData):
        pass

    @abstractmethod
    def _createResultData(self, trainedModels, evalDataList, testIndicesList, predictedVarNames) -> TCrossValData:
        pass

    def evalModel(self, model: VectorModel):
        trainedModels = [] if self.returnTrainedModels else None
        evalDataList = []
        testIndicesList = []
        predictedVarNames = None
        for evaluator in self.modelEvaluators:
            modelToFit: VectorModel = copy.deepcopy(model) if self.returnTrainedModels else model
            evaluator.fitModel(modelToFit)
            if predictedVarNames is None:
                predictedVarNames = modelToFit.getPredictedVariableNames()
            if self.returnTrainedModels:
                trainedModels.append(modelToFit)
            evalDataList.append(evaluator.evalModel(modelToFit))
            testIndicesList.append(evaluator.testData.outputs.index)
        return self._createResultData(trainedModels, evalDataList, testIndicesList, predictedVarNames)

    def computeMetrics(self, model: VectorModel):
        data = self.evalModel(model)
        return data.getEvalStatsCollection().aggStats()


class VectorRegressionModelCrossValidationData(PredictorModelCrossValidationData[VectorRegressionModel, VectorRegressionModelEvaluationData, RegressionEvalStats, RegressionEvalStatsCollection]):
    def _createEvalStatsCollection(self, l: List[RegressionEvalStats]) -> RegressionEvalStatsCollection:
        return RegressionEvalStatsCollection(l)


class VectorRegressionModelCrossValidator(VectorModelCrossValidator[VectorRegressionModelCrossValidationData]):
    # TODO: after switching to python3.8 we can move both methods to the base class by accessing the generic type at runtime
    def _createModelEvaluator(self, trainingData: InputOutputData, testData: InputOutputData) -> VectorRegressionModelEvaluator:
        return VectorRegressionModelEvaluator(trainingData, testData=testData, **self.evaluatorParams)

    def _createResultData(self, trainedModels, evalDataList, testIndicesList, predictedVarNames) -> VectorRegressionModelCrossValidationData:
        return VectorRegressionModelCrossValidationData(trainedModels, evalDataList, predictedVarNames, testIndicesList)


class VectorClassificationModelCrossValidationData(PredictorModelCrossValidationData[VectorClassificationModel, VectorClassificationModelEvaluationData, ClassificationEvalStats, ClassificationEvalStatsCollection]):
    def _createEvalStatsCollection(self, l: List[ClassificationEvalStats]) -> ClassificationEvalStatsCollection:
        return ClassificationEvalStatsCollection(l)


class VectorClassificationModelCrossValidator(VectorModelCrossValidator[VectorClassificationModelCrossValidationData]):
    def _createModelEvaluator(self, trainingData: InputOutputData, testData: InputOutputData):
        return VectorClassificationModelEvaluator(trainingData, testData=testData, **self.evaluatorParams)

    def _createResultData(self, trainedModels, evalDataList, testIndicesList, predictedVarNames) -> VectorClassificationModelCrossValidationData:
        return VectorClassificationModelCrossValidationData(trainedModels, evalDataList, predictedVarNames, testIndicesList)



import geopandas as gp
import networkx as nx
import numpy as np
import scipy
from itertools import combinations
from scipy.spatial.distance import euclidean
from shapely.geometry import MultiLineString
from typing import Callable, Dict

from .coordinates import extractCoordinatesArray, GeoDataFrameWrapper


def delaunayGraph(data: np.ndarray, edge_weight: Callable[[np.ndarray, np.ndarray], float] = euclidean):
    """
    The Delaunay triangulation of the data as networkx.Graph

    :param data:
    :param edge_weight: function to compute weight given two coordinate points
    :return: instance of networx.Graph where the edges contain additional datapoints entries for
        "weight" and for constants.COORDINATE_PAIR_KEY
    """
    tri = scipy.spatial.Delaunay(data)
    graph = nx.Graph()

    for simplex in tri.simplices:
        for vertex_id_pair in combinations(simplex, 2):
            coordinate_pair = tri.points[
                np.array(vertex_id_pair)]  # vertex_id_pair is a tuple and needs to be cast to an array
            graph.add_edge(*vertex_id_pair, weight=edge_weight(*coordinate_pair))
    return graph


class SpanningTree:
    """
    Wrapper around a tree-finding algorithm that will be applied on the Delaunay graph of the datapoints
    """
    def __init__(self, datapoints: np.ndarray, tree_finder: Callable[[nx.Graph], nx.Graph] = nx.minimum_spanning_tree):
        """
        :param datapoints:
        :param tree_finder: function mapping a graph to a subgraph. The default is minimum_spanning_tree
        """
        datapoints = extractCoordinatesArray(datapoints)
        self.tree = tree_finder(delaunayGraph(datapoints))
        edgeWeights = []
        self.coordinatePairs = []
        for edge in self.tree.edges.data():
            edgeCoordinateIndices, edgeData = [edge[0], edge[1]], edge[2]
            edgeWeights.append(edgeData["weight"])
            self.coordinatePairs.append(datapoints[edgeCoordinateIndices])
        self.edgeWeights = np.array(edgeWeights)

    def totalWeight(self):
        return self.edgeWeights.sum()

    def numEdges(self):
        return len(self.tree.edges)

    def meanEdgeWeight(self):
        return self.edgeWeights.mean()

    def summaryDict(self) -> Dict[str, float]:
        """
        Dictionary containing coarse information about the tree
        """
        return {
            "numEdges": self.numEdges(),
            "totalWeight": self.totalWeight(),
            "meanEdgeWeight": self.meanEdgeWeight()
        }


class CoordinateSpanningTree(SpanningTree, GeoDataFrameWrapper):
    """
    Wrapper around a tree-finding algorithm that will be applied on the Delaunay graph of the coordinates.
    Enhances the :class:`SpanningTree` class by adding methods and validation specific to geospatial coordinates.
    """
    def __init__(self, datapoints: np.ndarray, tree_finder: Callable[[nx.Graph], nx.Graph] = nx.minimum_spanning_tree):
        datapoints = extractCoordinatesArray(datapoints)
        super().__init__(datapoints, tree_finder=tree_finder)

    def multiLineString(self):
        return MultiLineString(self.coordinatePairs)

    def toGeoDF(self, crs='epsg:3857'):
        """
        :param crs: projection. By default pseudo-mercator
        :return: GeoDataFrame of length 1 with the tree as MultiLineString instance
        """
        gdf = gp.GeoDataFrame({"geometry": [self.multiLineString()]})
        gdf.crs = crs
        return gdf



import logging

import sklearn.ensemble
import sklearn.linear_model
import sklearn.neighbors
import sklearn.neural_network
import sklearn.svm

from .sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnMultiDimVectorRegressionModel


_log = logging.getLogger(__name__)


class SkLearnRandomForestVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, n_estimators=100, min_samples_leaf=10, random_state=42, **modelArgs):
        super().__init__(sklearn.ensemble.RandomForestRegressor,
            n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, random_state=random_state, **modelArgs)


class SkLearnLinearRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **modelArgs):
        super().__init__(sklearn.linear_model.LinearRegression, **modelArgs)


class SkLearnMultiLayerPerceptronVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self,
            random_state=42, hidden_layer_sizes=(100,), activation="relu", solver="adam",
            max_iter=200, **modelArgs):
        """
        :param random_state: the random seed
        :param hidden_layer_sizes: the sequence of hidden layer sizes
        :param activation: {‘identity’, ‘logistic’, ‘tanh’, ‘relu’} the activation function to use for hidden layers
            (the one used for the output layer is always 'identity')
        :param solver: {‘lbfgs’, ‘sgd’, ‘adam’} the name of the solver to apply
        :param max_iter: the number of iterations (gradient steps for L-BFGS, epochs for other solvers)
        :param modelArgs: additional arguments to pass on to MLPRegressor, see https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html
        """
        super().__init__(sklearn.neural_network.MLPRegressor,
            random_state=random_state, hidden_layer_sizes=hidden_layer_sizes, activation=activation, solver=solver, max_iter=max_iter,
            **modelArgs)


class SkLearnSVRVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **modelArgs):
        super().__init__(sklearn.svm.SVR, **modelArgs)


class SkLearnLinearSVRVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **modelArgs):
        super().__init__(sklearn.svm.LinearSVR, **modelArgs)


class SkLearnGradientBoostingVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, random_state=42, **modelArgs):
        super().__init__(sklearn.ensemble.GradientBoostingRegressor, random_state=random_state, **modelArgs)


class SkLearnKNeighborsVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **modelArgs):
        super().__init__(sklearn.neighbors.KNeighborsRegressor, **modelArgs)


class SkLearnExtraTreesVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, n_estimators=100, min_samples_leaf=10, random_state=42, **modelArgs):
        super().__init__(sklearn.ensemble.ExtraTreesRegressor,
            n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, random_state=random_state, **modelArgs)



import geopandas as gp
import logging
import numpy as np
from shapely.geometry import MultiPoint
from typing import Callable, Union, Iterable

from .base.clustering import ClusteringModel, SKLearnClustererProtocol, SKLearnClusteringModel
from ..util.cache import LoadSaveInterface
from ..util.coordinates import validateCoordinates, extractCoordinatesArray, TCoordinates, GeoDataFrameWrapper
from ..util.tracking import timed

log = logging.getLogger(__name__)


class CoordinateClusteringModel(ClusteringModel, GeoDataFrameWrapper):
    """
    Wrapper around a clustering model. This class adds additional, geospatial-specific features to the provided
    clusterer

    :param clusterer: an instance of ClusteringModel
    """
    def __init__(self, clusterer: ClusteringModel):
        self.clusterer = clusterer
        super().__init__(noiseLabel=clusterer.noiseLabel,
            maxClusterSize=clusterer.maxClusterSize, minClusterSize=clusterer.minClusterSize)

    class Cluster(ClusteringModel.Cluster, GeoDataFrameWrapper, LoadSaveInterface):
        """
        Wrapper around a coordinates array

        :param coordinates:
        :param identifier:
        """

        def __init__(self, coordinates: np.ndarray, identifier: Union[str, int]):
            validateCoordinates(coordinates)
            super().__init__(coordinates, identifier)

        def toGeoDF(self, crs='epsg:3857'):
            """
            Export the cluster as a GeoDataFrame of length 1 with the cluster as an instance of
            MultiPoint and the identifier as index.

            :param crs: projection. By default pseudo-mercator
            :return: GeoDataFrame
            """
            gdf = gp.GeoDataFrame({"geometry": [self.asMultipoint()]}, index=[self.identifier])
            gdf.index.name = "identifier"
            gdf.crs = crs
            return gdf

        def asMultipoint(self):
            """
            :return: The cluster's coordinates as a MultiPoint object
            """
            return MultiPoint(self.datapoints)

        @classmethod
        def load(cls, path):
            """
            Instantiate from a geopandas readable file containing a single row with an identifier and an instance
            of MultiPoint

            :param path:
            :return: instance of CoordinateCluster
            """
            log.info(f"Loading instance of {cls.__name__} from {path}")
            gdf = gp.read_file(path)
            if len(gdf) != 1:
                raise Exception(f"Expected {path} to contain a single row, instead got {len(gdf)}")
            identifier, multipoint = gdf.identifier.values[0], gdf.geometry.values[0]
            return cls(np.array([[p.x, p.y] for p in multipoint]), identifier)

        def save(self, path, crs="EPSG:3857"):
            """
            Saves the cluster's coordinates as shapefile

            :param crs:
            :param path:
            :return:
            """
            log.info(f"Saving instance of {self.__class__.__name__} as shapefile to {path}")
            self.toGeoDF(crs).to_file(path, index=True)

    def _computeLabels(self, x: np.ndarray) -> np.ndarray:
        validateCoordinates(x)
        return self.clusterer._computeLabels(x)

    def fit(self, coordinates: TCoordinates):
        """
        Fitting to coordinates from a numpy array, a MultiPoint object or a GeoDataFrame with one Point per row

        :param coordinates:
        :return:
        """
        coordinates = extractCoordinatesArray(coordinates)
        super().fit(coordinates)

    @timed
    def toGeoDF(self, condition: Callable[[Cluster], bool] = None, crs='epsg:3857',
            includeNoise=False) -> gp.GeoDataFrame:
        """
        GeoDataFrame containing all clusters found by the model.
        It is a concatenation of GeoDataFrames of individual clusters

        :param condition: if provided, only clusters fulfilling the condition will be included
        :param crs: projection. By default pseudo-mercator
        :param includeNoise:
        :return: GeoDataFrame with all clusters indexed by their identifier
        """
        geodf = gp.GeoDataFrame()
        geodf.crs = crs
        # TODO or not TODO: parallelize this or improve performance some another way
        for cluster in self.clusters(condition):
            geodf = geodf.append(cluster.toGeoDF(crs=crs))
        if includeNoise:
            geodf = geodf.append(self.noiseCluster().toGeoDF(crs=crs))
        return geodf

    def plot(self, includeNoise=False, condition=None, **kwargs):
        """
        Plots the resulting clusters with random coloring

        :param includeNoise: Whether to include the noise cluster
        :param condition: If provided, only clusters fulfilling this condition will be included
        :param kwargs: passed to GeoDataFrame.plot
        :return:
        """
        geodf = self.toGeoDF(condition=condition, includeNoise=includeNoise)
        geodf["color"] = np.random.random(len(geodf))
        if includeNoise:
            geodf.loc[self.noiseLabel, "color"] = 0
        geodf.plot(column="color", **kwargs)

    # the overriding of the following methods is only necessary for getting the type annotations right
    # if mypy ever permits annotating nested classes correctly, these methods can be removed
    def getCluster(self, clusterId: int) -> Cluster:
        return super().getCluster(clusterId)

    def noiseCluster(self) -> Cluster:
        return super().noiseCluster()

    def clusters(self, condition: Callable[[Cluster], bool] = None) -> Iterable[Cluster]:
        return super().clusters()


class SKLearnCoordinateClustering(CoordinateClusteringModel):
    """
    Wrapper around a sklearn clusterer. This class adds additional features like relabelling and convenient methods
    for handling geospatial data

    :param clusterer: a clusterer object compatible the sklearn API
    :param noiseLabel: label that is associated with the noise cluster or None
    :param minClusterSize: if not None, clusters below this size will be labeled as noise
    :param maxClusterSize: if not None, clusters above this size will be labeled as noise
    """
    def __init__(self, clusterer: SKLearnClustererProtocol, noiseLabel=-1,
                 minClusterSize: int = None, maxClusterSize: int = None):
        clusterer = SKLearnClusteringModel(clusterer, noiseLabel=noiseLabel,
                           minClusterSize=minClusterSize, maxClusterSize=maxClusterSize)
        super().__init__(clusterer)



"""
This module contains methods and classes that facilitate evaluation of different types of models. The suggested
workflow for evaluation is to use these higher-level functionalities instead of instantiating
the evaluation classes directly.
"""
# TODO: provide a notebook (and possibly an rst file) that illustrates standard evaluation scenarios and at the same
#  time serves as an integration test

import logging
from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Union, Generic, TypeVar, Optional, Sequence, Callable

import matplotlib.figure
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from .crossval import PredictorModelCrossValidationData, VectorRegressionModelCrossValidationData, \
    VectorClassificationModelCrossValidationData, \
    VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator
from .eval_stats.eval_stats_base import EvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats
from .eval_stats.eval_stats_regression import RegressionEvalStats
from .evaluator import VectorModelEvaluator, PredictorModelEvaluationData, VectorRegressionModelEvaluator, \
    VectorRegressionModelEvaluationData, VectorClassificationModelEvaluator, VectorClassificationModelEvaluationData
from ..data_ingest import InputOutputData
from ..util.io import ResultWriter
from ..vector_model import VectorClassificationModel, VectorRegressionModel, VectorModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=EvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)
TEvaluator = TypeVar("TEvaluator", bound=VectorModelEvaluator)
TEvalData = TypeVar("TEvalData", bound=PredictorModelEvaluationData)
TCrossValData = TypeVar("TCrossValData", bound=PredictorModelCrossValidationData)


def _isRegression(model: Optional[VectorModel], isRegression: Optional[bool]) -> bool:
    if model is None and isRegression is None or (model is not None and isRegression is not None):
        raise ValueError("One of the two parameters have to be passed: model or isRegression")

    if isRegression is None:
        model: VectorModel
        return model.isRegressionModel()
    return isRegression


def createVectorModelEvaluator(data: InputOutputData, model: VectorModel = None,
        isRegression: bool = None, **kwargs) \
            -> Union[VectorRegressionModelEvaluator, VectorClassificationModelEvaluator]:
    cons = VectorRegressionModelEvaluator if _isRegression(model, isRegression) else VectorClassificationModelEvaluator
    return cons(data, **kwargs)


def createVectorModelCrossValidator(data: InputOutputData, model: VectorModel = None,
        isRegression: bool = None, folds=5, **kwargs) \
            -> Union[VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator]:
    cons = VectorRegressionModelCrossValidator if _isRegression(model, isRegression) else VectorClassificationModelCrossValidator
    return cons(data, folds=folds, **kwargs)


def createEvaluationUtil(data: InputOutputData, model: VectorModel = None,
        isRegression: bool = None, evaluatorParams: Optional[Dict[str, Any]] = None,
        crossValidatorParams: Optional[Dict[str, Any]] = None) \
            -> Union["ClassificationEvaluationUtil", "RegressionEvaluationUtil"]:
    cons = RegressionEvaluationUtil if _isRegression(model, isRegression) else ClassificationEvaluationUtil
    return cons(data, evaluatorParams=evaluatorParams, crossValidatorParams=crossValidatorParams)


def evalModelViaEvaluator(model: TModel, inputOutputData: InputOutputData, testFraction=0.2,
        plotTargetDistribution=False, computeProbabilities=True, normalizePlots=True, randomSeed=60) -> TEvalData:
    """
    Evaluates the given model via a simple evaluation mechanism that uses a single split

    :param model: the model to evaluate
    :param inputOutputData: data on which to evaluate
    :param testFraction: the fraction of the data to test on
    :param plotTargetDistribution: whether to plot the target values distribution in the entire dataset
    :param computeProbabilities: only relevant if the model is a classifier
    :param normalizePlots: whether to normalize plotted distributions such that the sum/integrate to 1
    :param randomSeed:

    :return: the evaluation data
    """

    if plotTargetDistribution:
        title = "Distribution of target values in entire dataset"
        fig = plt.figure(title)

        outputDistributionSeries = inputOutputData.outputs.iloc[:, 0]
        log.info(f"Description of target column in training set: \n{outputDistributionSeries.describe()}")
        if not model.isRegressionModel():
            outputDistributionSeries = outputDistributionSeries.value_counts(normalize=normalizePlots)
            ax = sns.barplot(outputDistributionSeries.index, outputDistributionSeries.values)
            ax.set_ylabel("%")
        else:
            ax = sns.distplot(outputDistributionSeries)
            ax.set_ylabel("Probability density")
        ax.set_title(title)
        ax.set_xlabel("target value")
        fig.show()

    if model.isRegressionModel():
        evaluatorParams = dict(testFraction=testFraction, randomSeed=randomSeed)
    else:
        evaluatorParams = dict(testFraction=testFraction, computeProbabilities=computeProbabilities, randomSeed=randomSeed)
    ev = createEvaluationUtil(inputOutputData, model=model, evaluatorParams=evaluatorParams)
    return ev.performSimpleEvaluation(model, showPlots=True, logResults=True)


class EvaluationUtil(ABC, Generic[TModel, TEvaluator, TEvalData, TCrossValData, TEvalStats]):
    """
    Utility class for the evaluation of models based on a dataset
    """
    def __init__(self, inputOutputData: InputOutputData, evaluatorParams: Optional[Dict[str, Any]] = None,
            crossValidatorParams: Optional[Dict[str, Any]] = None):
        """
        :param inputOutputData: the data set to use for evaluation
        :param evaluatorParams: parameters with which to instantiate evaluators
        :param crossValidatorParams: parameters with which to instantiate cross-validators
        """
        if evaluatorParams is None:
            evaluatorParams = dict(testFraction=0.2)
        if crossValidatorParams is None:
            crossValidatorParams = dict(folds=5)
        self.evaluatorParams = evaluatorParams
        self.crossValidatorParams = crossValidatorParams
        self.inputOutputData = inputOutputData

    class ResultCollector:
        def __init__(self, showPlots: bool = True, resultWriter: Optional[ResultWriter] = None):
            self.showPlots = showPlots
            self.resultWriter = resultWriter

        def addFigure(self, name, fig: matplotlib.figure.Figure):
            if self.resultWriter is not None:
                self.resultWriter.writeFigure(name, fig, closeFigure=not self.showPlots)

        def child(self, addedFilenamePrefix):
            resultWriter = self.resultWriter
            if resultWriter:
                resultWriter = resultWriter.childWithAddedPrefix(addedFilenamePrefix)
            return self.__class__(showPlots=self.showPlots, resultWriter=resultWriter)

    def createEvaluator(self, model) -> TEvaluator:
        """
        Creates an evaluator which is suitable for evaluation of the given model
        :param model: the model for which to create an evaluator
        :return: an evaluator
        """
        return createVectorModelEvaluator(self.inputOutputData, model=model, **self.evaluatorParams)

    def performSimpleEvaluation(self, model: TModel, showPlots=False, logResults=True, resultWriter: ResultWriter = None,
            additionalEvaluationOnTrainingData=False) -> TEvalData:
        evaluator = self.createEvaluator(model)
        evaluator.fitModel(model)

        def gatherResults(evalResultData, resultWriter, subtitlePrefix=""):
            strEvalResults = ""
            for predictedVarName in model.getPredictedVariableNames():
                strEvalResult = str(evalResultData.getEvalStats(predictedVarName))
                if logResults:
                    log.info(f"Evaluation results for {predictedVarName}: {strEvalResult}")
                strEvalResults += predictedVarName + ": " + strEvalResult + "\n"
            if resultWriter is not None:
                resultWriter.writeTextFile("evaluator-results", strEvalResults)
            self.createPlots(evalResultData, showPlots=showPlots, resultWriter=resultWriter, subtitlePrefix=subtitlePrefix)

        evalResultData = evaluator.evalModel(model)
        gatherResults(evalResultData, resultWriter)
        if additionalEvaluationOnTrainingData:
            evalResultDataTrain = evaluator.evalModel(model, onTrainingData=True)
            gatherResults(evalResultDataTrain, resultWriter.childWithAddedPrefix("-onTrain-"), subtitlePrefix="[onTrain] ")

        return evalResultData

    @staticmethod
    def _resultWriterForModel(resultWriter: Optional[ResultWriter], model: TModel) -> Optional[ResultWriter]:
        if resultWriter is None:
            return None
        return resultWriter.childWithAddedPrefix(model.getName() + "-")

    def performCrossValidation(self, model: TModel, showPlots=False, logResults=True, resultWriter: Optional[ResultWriter] = None) -> TCrossValData:
        """
        Evaluates the given model via cross-validation

        :param model: the model to evaluate
        :param showPlots: whether to show plots that visualise evaluation results (combining all folds)
        :param logResults: whether to log evaluation results
        :param resultWriter: a writer with which to store text files and plots. The evaluated model's name is added to each filename
            automatically
        :return: cross-validation result data
        """
        resultWriter = self._resultWriterForModel(resultWriter, model)
        crossValidator = createVectorModelCrossValidator(self.inputOutputData, model=model, **self.crossValidatorParams)
        crossValidationData = crossValidator.evalModel(model)
        strEvalResults = str(crossValidationData.getEvalStatsCollection().aggStats())
        if logResults:
            log.info(f"Cross-validation results: {strEvalResults}")
        if resultWriter is not None:
            resultWriter.writeTextFile("evaluator-results", strEvalResults)
        self.createPlots(crossValidationData, showPlots=showPlots, resultWriter=resultWriter)
        return crossValidationData

    def compareModelsCrossValidation(self, models: Sequence[TModel], resultWriter: Optional[ResultWriter] = None) -> pd.DataFrame:
        """
        Compares several models via cross-validation

        :param models: the models to compare
        :param resultWriter: a writer with which to store results of the comparison
        :return: a data frame containing evaluation metrics on all models
        """
        statsList = []
        for model in models:
            crossValidationResult = self.performCrossValidation(model, resultWriter=resultWriter)
            stats = crossValidationResult.getEvalStatsCollection().aggStats()
            stats["modelName"] = model.getName()
            statsList.append(stats)
        resultsDF = pd.DataFrame(statsList).set_index("modelName")
        strResults = f"Model comparison results:\n{resultsDF.to_string()}"
        log.info(strResults)
        if resultWriter is not None:
            resultWriter.writeTextFile("model-comparison-results", strResults)
        return resultsDF

    # TODO: a parameter predictedVarName seemed to have been present at some point but has now disappeared,
    #   should we resuscitate it?
    def createPlots(self, data: Union[TEvalData, TCrossValData], showPlots=True, resultWriter: Optional[ResultWriter] = None, subtitlePrefix: str = ""):
        """
        Creates default plots that visualise the results in the given evaluation data

        :param data: the evaluation data for which to create the default plots
        :param showPlots: whether to show plots
        :param resultWriter: if not None, plots will be written using this writer
        :param subtitlePrefix: a prefix to add to the subtitle (which itself is the model name)
        """
        if not showPlots and resultWriter is None:
            return
        resultCollector = self.ResultCollector(showPlots=showPlots, resultWriter=resultWriter)
        self._createPlots(data, resultCollector, subtitle=subtitlePrefix + data.modelName)

    def _createPlots(self, data: Union[TEvalData, TCrossValData], resultCollector: ResultCollector, subtitle=None):

        def createPlots(predVarName, rc, subt):
            if isinstance(data, PredictorModelCrossValidationData):
                evalStats = data.getEvalStatsCollection(predictedVarName=predVarName).getGlobalStats()
            elif isinstance(data, PredictorModelEvaluationData):
                evalStats = data.getEvalStats(predictedVarName=predVarName)
            else:
                raise ValueError(f"Unexpected argument: data={data}")
            return self._createEvalStatsPlots(evalStats, rc, subtitle=subt)

        predictedVarNames = data.predictedVarNames
        if len(predictedVarNames) == 1:
            createPlots(predictedVarNames[0], resultCollector, subtitle)
        else:
            for predictedVarName in predictedVarNames:
                createPlots(predictedVarName, resultCollector.child(predictedVarName+"-"), f"{predictedVarName}, {subtitle}")

    @abstractmethod
    def _createEvalStatsPlots(self, evalStats: TEvalStats, resultCollector: ResultCollector, subtitle=None):
        """
        :param evalStats: the evaluation results for which to create plots
        :param resultCollector: the collector to which all plots are to be passed
        :param subtitle: the subtitle to use for generated plots (if any)
        """
        pass


class RegressionEvaluationUtil(EvaluationUtil[VectorRegressionModel, VectorRegressionModelEvaluator, VectorRegressionModelEvaluationData, VectorRegressionModelCrossValidationData, RegressionEvalStats]):
    def _createEvalStatsPlots(self, evalStats: RegressionEvalStats, resultCollector: EvaluationUtil.ResultCollector, subtitle=None):
        resultCollector.addFigure("error-dist", evalStats.plotErrorDistribution(titleAdd=subtitle))
        resultCollector.addFigure("heatmap-gt-pred", evalStats.plotHeatmapGroundTruthPredictions(titleAdd=subtitle))
        resultCollector.addFigure("scatter-gt-pred", evalStats.plotScatterGroundTruthPredictions(titleAdd=subtitle))


class ClassificationEvaluationUtil(EvaluationUtil[VectorClassificationModel, VectorClassificationModelEvaluator, VectorClassificationModelEvaluationData, VectorClassificationModelCrossValidationData, ClassificationEvalStats]):
    def _createEvalStatsPlots(self, evalStats: ClassificationEvalStats, resultCollector: EvaluationUtil.ResultCollector, subtitle=None):
        resultCollector.addFigure("confusion-matrix", evalStats.plotConfusionMatrix(titleAdd=subtitle))


class MultiDataEvaluationUtil:
    def __init__(self, inputOutputDataDict: Dict[str, InputOutputData], keyName: str = "dataset"):
        """
        :param inputOutputDataDict: a dictionary mapping from names to the data sets with which to evaluate models
        :param keyName: a name for the key value used in inputOutputDataDict
        """
        self.inputOutputDataDict = inputOutputDataDict
        self.keyName = keyName

    def compareModelsCrossValidation(self, isRegression, modelFactories: Sequence[Callable[[], VectorModel]],
            resultWriter: Optional[ResultWriter] = None, writePerDatasetResults=True,
            crossValidatorParams: Optional[Dict[str, Any]] = None, columnNameForModelRanking: str = None, rankMax=True) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        :param isRegression: flag indicating whether the models to evaluate are regression models
        :param modelFactories: a sequence of factory functions for the creation of models to evaluate
        :param resultWriter: a writer with which to store results
        :param writePerDatasetResults: whether to use resultWriter (if not None) in order to generate detailed results for each
            dataset in a subdirectory named according to the name of the dataset
        :param crossValidatorParams: parameters to use for the instantiation of cross-validators
        :param columnNameForModelRanking: column name to use for ranking models
        :param rankMax: if true, use max for ranking, else min
        :return: a pair of data frames (allDF, meanDF) where allDF contains all the individual cross-validation results
            for every dataset and meanDF contains one row for each model with results averaged across datasets
        """
        allResults = pd.DataFrame()
        for key, inputOutputData in self.inputOutputDataDict.items():
            log.info(f"Evaluating models for {key}")
            ev = createEvaluationUtil(inputOutputData, isRegression=isRegression, crossValidatorParams=crossValidatorParams)
            models = [f() for f in modelFactories]
            childResultWriter = resultWriter.childForSubdirectory(key) if writePerDatasetResults else None
            df = ev.compareModelsCrossValidation(models, resultWriter=childResultWriter)
            df[self.keyName] = key
            df["modelName"] = df.index
            if columnNameForModelRanking is not None:
                if columnNameForModelRanking not in df.columns:
                    raise ValueError(f"Rank metric {columnNameForModelRanking} not contained in columns {df.columns}")
                df["bestModel"] = 0
                if rankMax:
                    df["bestModel"].loc[df[columnNameForModelRanking].idxmax()] = 1
                else:
                    df["bestModel"].loc[df[columnNameForModelRanking].idxmin()] = 1
            df = df.reset_index(drop=True)
            allResults = pd.concat((allResults, df))
        strAllResults = f"All results:\n{allResults.to_string()}"
        log.info(strAllResults)
        meanResults = allResults.groupby("modelName").mean()
        strMeanResults = f"Mean results:\n{meanResults.to_string()}"
        log.info(strMeanResults)
        if resultWriter is not None:
            resultWriter.writeTextFile("model-comparison-results", strMeanResults + "\n\n" + strAllResults)
        return allResults, meanResults



from .crossval import VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator, \
    VectorClassificationModelCrossValidationData, VectorRegressionModelCrossValidationData
from .eval_util import RegressionEvaluationUtil, ClassificationEvaluationUtil, MultiDataEvaluationUtil, \
    evalModelViaEvaluator
from .evaluator import VectorClassificationModelEvaluator, VectorRegressionModelEvaluator, \
    VectorRegressionModelEvaluationData, VectorClassificationModelEvaluationData



import collections
import logging
import math
import random
import time
from abc import ABC, abstractmethod
from typing import Optional, Tuple, Callable, Type, Sequence, TypeVar, Generic

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

log = logging.getLogger(__name__)


class SATemperatureSchedule(ABC):
    """
    Describes how temperature changes as the annealing process goes on.
    The function maps a degree of completion (in the interval from 0 to 1) to a temperature value.
    """

    @abstractmethod
    def temperature(self, degreeOfCompletion):
        """
        Computes the temperature for a degree of completion in [0, 1]
        :param degreeOfCompletion: the degree to which the simulated annealing process has completed in [0, 1]
        :return: the temperature
        """
        pass

    def probability(self, degreeOfCompletion, costDelta):
        T = self.temperature(degreeOfCompletion)
        p = math.exp(-costDelta / T) if T > 0.0 else 0.0
        return p, T

    @abstractmethod
    def _getParams(self):
        pass

    def __str__(self):
        return f"{self.__class__.__name__}{self._getParams()}"


class SATemperatureScheduleFixed(SATemperatureSchedule):
    """A schedule with a constant temperature"""
    def __init__(self, t):
        self.t = t

    def temperature(self, degreeOfCompletion):
        return self.t

    def _getParams(self):
        return {"fixedTemp": self.t}


class SATemperatureScheduleExponential(SATemperatureSchedule):
    """ A temperature schedule for simulated annealing where the temperature drops exponentially. """

    def __init__(self, t0, t1, exponentFactor):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param exponentFactor: factor with which to multiply the exponent (the larger the factor, the faster the temperature drops)
        """
        super().__init__()
        if t1 > t0 or t0 <= 0:
            raise ValueError("Inadmissible temperatures given.")

        if exponentFactor < 1:
            raise ValueError("The exponent factor cannot be less than 1.")

        self.t0 = t0
        self.t1 = t1
        self.tDrop = t0 - t1
        self.exponentFactor = exponentFactor

    def temperature(self, degreeOfCompletion):
        return self.t0 - self.tDrop * (1.0 - math.exp(-self.exponentFactor * degreeOfCompletion))

    def _getParams(self):
        return {"t0": self.t0, "t1": self.t1, "expFactor": self.exponentFactor}


def reverseSigmoid(degreeOfCompletion, v0, v1, steepness, midDegree=0.5):
    return v0 - (v0-v1) / (1.0 + math.exp(-steepness * (degreeOfCompletion - midDegree)))


class SATemperatureScheduleReverseSigmoid(SATemperatureSchedule):
    """A temperature schedule for simulated annealing where the temperature drops in a reverse sigmoid curve (based on the logistic function)"""

    def __init__(self, t0, t1, steepness, midDegree=0.5):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param steepness: Factor which roughly corresponds to the (negative) slope at the inflection point 0.5.
            For the sigmoid shape to be sufficiently pronounced, a value of at least 5 is recommended (the closer the value is to 1,
            the more the curve approaches a linear decay).
        :param midDegree: the degree of completion at which the function shall return the temperature (t0+t1)/2
        """
        super().__init__()
        if t1 > t0 or t0 <= 0:
            raise ValueError("Inadmissible temperatures given.")
        if midDegree < 0 or midDegree > 1:
            raise Exception("Mid-degree must be between 0 and 1")

        self.t0 = t0
        self.t1 = t1
        self.steepness = steepness
        self.midDegree = midDegree

    def temperature(self, degreeOfCompletion):
        return reverseSigmoid(degreeOfCompletion, self.t0, self.t1, self.steepness, self.midDegree)

    def _getParams(self):
        return {"t0": self.t0, "t1": self.t1, "steepness": self.steepness, "midDegree": self.midDegree}


class SATemperatureScheduleReverseSigmoidSymmetric(SATemperatureScheduleReverseSigmoid):
    """
    A variant of the logistic schedule with a reverse sigmoid shape, where the probability of acceptance for a given assumed positive
    cost delta is symmetric. "Symmetric" here means that half the schedule is to be above 0.5 and half of it below 0.5.
    """
    def __init__(self, t0, t1, steepness, costDeltaForSymmetry):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param steepness: Factor which roughly corresponds to the (negative) slope at the inflection point 0.5.
            For the sigmoid shape to be sufficiently pronounced, a value of at least 8 is recommended (the closer the value is to 1,
            the more the curve approaches a linear decay).
        :param costDeltaForSymmetry: the (positive) cost delta value which the curve shall be "symmetric"
        """
        k = steepness
        cdelta = costDeltaForSymmetry
        midDegree = -(2*math.log(-(13614799*t0-19642003*cdelta)/(13614799*t1-19642003*cdelta))-k)/(2*k)
        super().__init__(t0, t1, steepness, midDegree)


class SATemperatureSchedulePower(SATemperatureSchedule):
    """
    A temperature schedule for simulated annealing where the temperature drops with powers of the degree of completion d,
    i.e. the temperature drop is proportional to d to the power of e for a given exponent e.
    """

    def __init__(self, t0, t1, exponent):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param exponent: the exponent of the power function
        """
        super().__init__()
        if t1 > t0 or t0 <= 0:
            raise ValueError("Inadmissible temperatures given.")

        self.t0 = t0
        self.t1 = t1
        self.tDrop = t0 - t1
        self.exponent = exponent

    def temperature(self, degreeOfCompletion):
        return self.t0 - self.tDrop * (math.pow(degreeOfCompletion, self.exponent))

    def _getParams(self):
        return {"t0": self.t0, "t1": self.t1, "exponent": self.exponent}


class SAProbabilityFunction(ABC):
    def __init__(self, **paramsDict):
        self.paramsDict = paramsDict

    def __str__(self):
        return f"{self.__class__.__name__}{self.paramsDict}"

    @abstractmethod
    def __call__(self, degreeOfCompletion):
        pass


class SAProbabilityFunctionLinear(SAProbabilityFunction):
    """A probability function where probabilities decay linearly"""
    def __init__(self, p0=1.0, p1=0.0):
        super().__init__(p0=p0, p1=p1)
        self.p0 = p0
        self.p1 = p1

    def __call__(self, degreeOfCompletion):
        return self.p0 - degreeOfCompletion * (self.p0 - self.p1)


class SAProbabilityFunctionReverseSigmoid(SAProbabilityFunction):
    """A probability function where probabilities decay in a reverse sigmoid shape"""
    def __init__(self, p0=1.0, p1=0.0, steepness=10):
        """
        :param p0: the probability at the beginning (degree of completion 0)
        :param p1: the probability at the end (degree of completion 1)
        :param steepness: the steepness of the sigmoid curve
        """
        super().__init__(p0=p0, p1=p1, steepness=steepness)
        self.p0 = p0
        self.p1 = p1
        self.steepness = steepness

    def __call__(self, degreeOfCompletion):
        return reverseSigmoid(degreeOfCompletion, self.p0, self.p1, self.steepness)


class SAProbabilityFunctionConstant(SAProbabilityFunction):
    """A constant probability function (which returns the same probability for all degrees of completion)"""
    def __init__(self, p):
        super().__init__(p=p)
        self.p = p

    def __call__(self, degreeOfCompletion):
        return self.p


class SAProbabilitySchedule(SATemperatureSchedule):
    """
    A temperature schedule where temperatures are derived from a probability schedule that is to apply to a reference cost delta, which
    is either given or is computed from observed values (the latter resulting in an adaptive schedule).
    It converts a function that returns probabilities for degrees of completion into a corresponding temperature schedule.
    """
    def __init__(self, referenceCostDelta: Optional[float], probabilityFunction: SAProbabilityFunction):
        """
        Creates a temperature schedule for a reference cost delta (which can also be computed automatically from observed data)
        and probability function:
        The schedule will return temperatures such that for referenceCostDelta, the probability of acceptance of a move at
        degree of completion d in [0,1] will be probabilityFunction(d).

        :param referenceCostDelta: the (positive) cost delta for which the probability function is to apply;
            if None, adaptively determine it from the empirical mean
        :param probabilityFunction: a function which maps degrees of completion in [0,1] to probabilities in [0,1]
        """
        self.adaptive = referenceCostDelta is None
        self.referenceCostDelta = referenceCostDelta
        self.probabilityFunction = probabilityFunction
        self.paramsDict = {
            "refCostDelta": referenceCostDelta if referenceCostDelta is not None else "adaptive",
            "probabilityFunction": str(probabilityFunction)}

        # helper variables for adaptive mode
        self._costDeltaSum = 0
        self._costDeltaCount = 0

    def temperature(self, degreeOfCompletion):
        if self.adaptive and self._costDeltaCount == 0:
            raise Exception("Cannot generate a temperature from an adaptive schedule without any previous cost-delta samples")
        p = self.probabilityFunction(degreeOfCompletion)
        if p == 0.0:
            return 0
        else:
            return -self.referenceCostDelta / math.log(p)

    def probability(self, degreeOfCompletion, costDelta):
        if self.adaptive:
            self._costDeltaSum += costDelta
            self._costDeltaCount += 1
            self.referenceCostDelta = max(self._costDeltaSum / self._costDeltaCount, 1e-10)
        return super().probability(degreeOfCompletion, costDelta)

    def _getParams(self):
        return self.paramsDict


class SACostValue(ABC):
    """Representation of an immutable cost value"""

    @abstractmethod
    def value(self):
        """Returns the numeric cost value"""
        pass

    @abstractmethod
    def add(self, other) -> 'SACostValue':
        pass


class SACostValueNumeric(SACostValue):
    def __init__(self, scalar):
        self._value = scalar

    def __str__(self):
        return str(self.value())

    def value(self):
        return self._value

    def add(self, other: 'SACostValueNumeric') -> 'SACostValueNumeric':
        return SACostValueNumeric(self._value + other._value)


class SAState(ABC):
    """Represents the state/variable assignment during a simulated annealing process"""

    def __init__(self, r: random.Random):
        self.r = r
        self.cost = self.computeCostValue()

    @abstractmethod
    def computeCostValue(self) -> SACostValue:
        """Computes the costs of this state (from scratch)"""
        pass

    @abstractmethod
    def getStateRepresentation(self):
        """
        Returns a compact state representation (for the purpose of archiving a hitherto best result), which can later be
        written to output data structures via applyStateRepresentation.

        :return: a compact state representation of some sort
        """
        pass

    @abstractmethod
    def applyStateRepresentation(self, representation):
        """
        Writes the given state representation to a suitable object given at construction.
        This function serves to write the results of the optimisation.

        :param representation: a representation as returned by getStateRepresentation
        """
        pass


TSAState = TypeVar("TSAState", bound=SAState)


class SAOperator(Generic[TSAState]):
    """
    An operator which, when applied with appropriately chosen parameters, can transform a state into another
    state during simulated annealing
    """

    def __init__(self, state: TSAState):
        """
        :param state: the state to which the operator is applied
        """
        self.state = state

    def applyCostChange(self, costDelta: SACostValue):
        """
        Applies the cost change to the state given at construction

        :param costDelta: the cost change to apply
        """
        self.state.cost = self.state.cost.add(costDelta)

    @abstractmethod
    def applyStateChange(self, *params):
        """
        Applies the operator to the state, i.e. it makes the changes to the state only
        (and does not update the associated costs)

        :param params: the parameters with which the operator is to be applied
        :return:
        """
        pass

    def apply(self, params: Tuple, costDelta: SACostValue):
        """
        Applies the operator to the state given at construction, changing the state and updating the costs appropriately

        :param params: the parameters with which the operator is to be applied
        :param costDelta: the cost change that results from the application
        :return:
        """
        self.applyCostChange(costDelta)
        self.applyStateChange(*params)

    @abstractmethod
    def costDelta(self, *params) -> SACostValue:
        """
        Computes the cost change that would apply when applying this operator with the given parameters

        :param params: an arbitrary list of parameters (specific to the concrete operator)
        :return:
        """
        pass

    @abstractmethod
    def chooseParams(self) -> Optional[Tuple[Tuple, Optional[SACostValue]]]:
        """
        Chooses parameters for the application of the operator (e.g. randomly or greedily).

        :return: a tuple (params, costValue) or None if no suitable parameters are found, where params is the list of chosen
            parameters and costValue is either an instance of CostValue or None if the costs have not been computed.
        """
        pass


class RelativeFrequencyCounter:
    def __init__(self):
        self.numTotal = 0
        self.numRelevant = 0

    def count(self, isRelevantEvent):
        self.numTotal += 1
        if isRelevantEvent:
            self.numRelevant += 1

    def __str__(self):
        info = f"{self.numRelevant}/{self.numTotal}"
        if self.numTotal > 0:
            info += f", {100 * self.numRelevant / self.numTotal:.2f}%"
        return f"RelativeFrequencyCounter[{info}]"


class SAChain:
    """Manages the progression of one state during simulated annealing"""

    log = log.getChild(__qualname__)

    def __init__(self, stateFactory: Callable[[random.Random], SAState], schedule: SATemperatureSchedule,
            opsAndWeights: Sequence[Tuple[Callable[[SAState], SAOperator], float]], randomSeed, collectStats=False):
        self.schedule = schedule
        self.r = random.Random(randomSeed)
        self.state = stateFactory(self.r)
        self.collectStats = collectStats
        operators, weights = zip(*opsAndWeights)
        cumWeights, s = [], 0
        for weight in weights:
            s += weight
            cumWeights.append(s)
        self.ops = [cons(self.state) for cons in operators]
        self.opCumWeights = cumWeights
        self.stepsTaken = 0
        self.countNoneParams = 0
        self.countBestUpdates = -1
        self.bestCost = None
        self.bestStateRepr = None
        self.loggedSeries = collections.defaultdict(lambda: [])
        self._updateBestState()

        if self.collectStats:
            self.operatorInapplicabilityCounters = {}
            for op in self.ops:
                self.operatorInapplicabilityCounters[op] = RelativeFrequencyCounter()

    def _updateBestState(self):
        cost = self.state.cost
        if self.bestCost is None or cost.value() < self.bestCost.value():
            self.bestCost = cost
            self.bestStateRepr = self.state.getStateRepresentation()
            self.countBestUpdates += 1

    def step(self, degreeOfCompletion):
        r = self.r

        # make move
        op = r.choices(self.ops, cum_weights=self.opCumWeights, k=1)[0]
        paramChoice = op.chooseParams()
        if paramChoice is None:
            self.countNoneParams += 1
        else:
            params, costChange = paramChoice
            if costChange is None:
                costChange = op.costDelta(*params)
            if costChange.value() < 0:
                makeMove = True
            else:
                costChangeValue = costChange.value()
                p, T = self.schedule.probability(degreeOfCompletion, costChangeValue)
                makeMove = r.random() <= p
                self.log.debug(f'p: {p}, T: {T}, costDelta: {costChangeValue}, move: {makeMove}')
                if self.collectStats:
                    self.loggedSeries["temperatures"].append(T)
                    self.loggedSeries["probabilities"].append(p)
            if makeMove:
                op.apply(params, costChange)
                self._updateBestState()
            if self.collectStats:
                self.loggedSeries["costDeltas"].append(costChange.value())
        if self.collectStats:
            self.loggedSeries["bestCostValues"].append(self.bestCost.value())
            self.loggedSeries["costValues"].append(self.state.cost.value())
            self.operatorInapplicabilityCounters[op].count(paramChoice is None)

        self.stepsTaken += 1

        if self.log.isEnabledFor(logging.DEBUG):
            self.log.debug(f"Step {self.stepsTaken}: cost={self.state.cost}; best cost={self.bestCost}")

    def logStats(self):
        stats = {"useless moves total (None params)": f"{self.countNoneParams}/{self.stepsTaken}"}
        if self.collectStats:
            for op, counter in self.operatorInapplicabilityCounters.items():
                stats[f"useless moves of {op}"] = str(counter)
            loggedCostDeltas = self.loggedSeries["costDeltas"]
            if loggedCostDeltas:
                stats["mean cost delta"] = f"{np.mean(loggedCostDeltas):.3f} +- { np.std(loggedCostDeltas):.3f}"
                absCostDeltas = np.abs(loggedCostDeltas)
                stats["mean absolute cost delta"] = f"{np.mean(absCostDeltas):.3f} +- {np.std(absCostDeltas):.3f}"
                positiveCostDeltas = [cd for cd in loggedCostDeltas if cd > 0]
                if positiveCostDeltas:
                    stats["positive cost delta"] = f"mean={np.mean(positiveCostDeltas):.3f} +- {np.std(positiveCostDeltas):.3f}," \
                                                   f" max={np.max(positiveCostDeltas):.3f}"
        statsJoin = "\n    " if self.collectStats else "; "
        self.log.info(f"Stats: {statsJoin.join([key + ': ' + value for (key, value) in stats.items()])}")
        self.log.info(f"Best solution has {self.bestCost} after {self.countBestUpdates} updates of best state")

    def applyBestState(self):
        """Writes the best state found in this chain to the result object"""
        self.state.applyStateRepresentation(self.bestStateRepr)

    def plotSeries(self, seriesName):
        """
        Plots one of the logged series

        :param seriesName: the name of the series (see getSeries)
        """
        series = self.getSeries(seriesName)
        plt.figure()
        series.plot(title=seriesName)

    def getSeries(self, seriesName):
        """
        Gets one of the logged series (for collectStats==True)

        :param seriesName: name of the series: one of "temperatures", "probabilities", "costDeltas", "bestCostValues", "costValues
        """
        if not self.collectStats:
            raise Exception("No stats were collected")
        if seriesName not in self.loggedSeries:
            raise Exception("Unknown series")
        return pd.Series(self.loggedSeries[seriesName])


class SimulatedAnnealing:
    log = log.getChild(__qualname__)

    """
    The simulated annealing algorithm for discrete optimisation (cost minimisation)
    """
    def __init__(self, scheduleFactory: Callable[[], SATemperatureSchedule], opsAndWeights: Sequence[Tuple[Callable[[SAState], SAOperator], float]],
            maxSteps: int = None, duration: float = None, randomSeed=42, collectStats=False):
        """
        :param scheduleFactory: a factory for the creation of the temperature schedule for the annealing process
        :param opsAndWeights: a list of operators with associated weights (which are to indicate the non-normalised probability of chosing the associated operator)
        :param maxSteps: the number of steps for which to run the optimisation; may be None (if not given, duration must be provided)
        :param duration: the duration, in seconds, for which to run the optimisation; may be None (if not given, maxSteps must be provided)
        :param randomSeed: the random seed to use for all random choices
        :param collectStats: flag indicating whether to collect additional statics which will be logged
        """
        if maxSteps is not None and maxSteps <= 0:
            raise ValueError("The number of iterations should be greater than 0.")
        if maxSteps is None and duration is None or (maxSteps is not None and duration is not None):
            raise ValueError("Exactly one of {maxSteps, duration} must be specified.")
        if duration is not None and duration <= 0:
            raise ValueError("Duration must be greater than 0 if provided")
        self.scheduleFactory = scheduleFactory
        self.maxSteps = maxSteps
        self.duration = duration
        self.randomSeed = randomSeed
        self.opsAndWeights = opsAndWeights
        self.collectStats = collectStats
        self._chain = None

    def optimise(self, stateFactory: Callable[[random.Random], SAState]):
        """
        Applies the annealing process starting with a state created via the given factory.
        The result of the optimisation (i.e. the final best state representation) is written via the state's
        applyStateRepresentation method, which should write to an object the state receives at construction.

        :param stateFactory: the factory with which to create the (initial) state
        """
        chain = SAChain(stateFactory, self.scheduleFactory(), opsAndWeights=self.opsAndWeights, randomSeed=self.randomSeed, collectStats=self.collectStats)
        self.log.info(f"Running simulated annealing with {len(self.opsAndWeights)} operators for {'%d steps' % self.maxSteps if self.maxSteps is not None else '%d seconds' % self.duration} ...")
        startTime = time.time()
        while True:
            timeElapsed = time.time() - startTime
            if (self.maxSteps is not None and chain.stepsTaken >= self.maxSteps) or (self.duration is not None and timeElapsed >= self.duration):
                break
            if self.maxSteps is not None:
                degreeOfCompletion = chain.stepsTaken / self.maxSteps
            else:
                degreeOfCompletion = timeElapsed / self.duration
            chain.step(degreeOfCompletion)
        self.log.info(f"Simulated annealing completed after {time.time() - startTime:.1f} seconds, {chain.stepsTaken} steps")
        chain.logStats()
        chain.applyBestState()
        if self.collectStats:
            self._chain = chain

    def getChain(self) -> Optional[SAChain]:
        """
        Gets the chain used by the most recently completed application (optimise call) of this object
        for the case where stats collection was enabled; the chain then contains relevant series and may be used
        to generate plots. If stats collection was not enabled, returns None.
        """
        return self._chain


class ParallelTempering:
    log = log.getChild(__qualname__)

    """
    The parallel tempering algorithm for discrete optimisation (cost minimisation)
    """
    def __init__(self, numChains, opsAndWeights: Sequence[Tuple[Type[SAOperator], float]],
                 schedule: SATemperatureSchedule = None, probabilityFunction: SAProbabilityFunction = None,
                 maxSteps: int = None, duration: float = None, randomSeed=42, logCostProgression=False):
        """
        Creates a parallel tempering optimiser with the given number of chains and operators for each chain.
        To determine the schedule to use for each chain, either schedule or probabilityFunction must be provided.
        It is usually more robust to use adaptive schedules and therefore to provide probabilityFunction.

        :param numChains: the number of parallel chains
        :param opsAndWeights: a list of operators with associated weights (which are to indicate the non-normalised probability of chosing the associated operator)
        :param schedule: the temperature schedule from which numChains temperatures of chains are drawn (using equidistant degrees of completion); if None, must provide probabilityFunction
        :param probabilityFunction: the probability function from which numChains probabilities for adaptive probability schedules, each using a constant probability, are to be drawn; if None, must provide schedule
        :param maxSteps: the number of steps for which to run the optimisation; may be None (if not given, duration must be provided)
        :param duration: the duration, in seconds, for which to run the optimisation; may be None (if not given, maxSteps must be provided)
        :param randomSeed: the random seed to use for all random choices
        :param logCostProgression: whether to log cost progression of all chains (such that it can be plotted after the fact via plotCostProgression)
        """
        if maxSteps is not None and maxSteps <= 0:
            raise ValueError("The number of iterations should be greater than 0.")
        if (maxSteps is None and duration is None) or (maxSteps is not None and duration is not None):
            raise ValueError("Exactly one of {maxSteps, duration} must be specified.")
        if duration is not None and duration <= 0:
            raise ValueError("duration should be greater than 0 if provided.")
        if numChains < 2:
            raise ValueError("Number of chains must be at least 2.")
        if (schedule is None and probabilityFunction is None) or (schedule is not None and probabilityFunction is not None):
            raise ValueError("Exactly one of {schedule, probabilityFunction} must be given")
        self.maxSteps = maxSteps
        self.duration = duration
        self.randomSeed = randomSeed
        self.numChains = numChains
        self.baseSchedule = schedule
        self.baseProbabilityFunction = probabilityFunction
        self.opsAndWeights = opsAndWeights
        self.logCostProgression = logCostProgression

        # transient members
        self._costProgressions = None
        self._scheduleParamStrings = None

    def _createSchedules(self):
        degreeStep = 1.0 / (self.numChains-1)
        degreesOfCompletion = [i*degreeStep for i in range(self.numChains)]
        if self.baseSchedule is not None:
            # create schedules with fixed temperatures taken from base schedule
            temperatures = [self.baseSchedule.temperature(d) for d in degreesOfCompletion]
            self._scheduleParamStrings = ["T=%.2f" % t for t in temperatures]
            return [SATemperatureScheduleFixed(t) for t in temperatures]
        else:
            # create adaptive probability schedules based on probabilities taken from base probability function
            probabilities = [self.baseProbabilityFunction(d) for d in degreesOfCompletion]
            self._scheduleParamStrings = ["p=%.3f" % p for p in probabilities]
            return [SAProbabilitySchedule(None, SAProbabilityFunctionConstant(p)) for p in probabilities]

    def optimise(self, stateFactory: Callable[[random.Random], SAState]):
        """
        Applies the optimisation process starting, in each chain, with a state created via the given factory.
        The result of the optimisation (i.e. the final best state representation) is written by calling the
        applyStateRepresentation method on one of the states, which should write to a suitable object each
        state receives at construction.

        :param stateFactory: the factory with which to create the states for all chains
        """
        self.log.info(f"Running parallel tempering with {self.numChains} chains, {len(self.opsAndWeights)} operators for {'%d steps' % self.maxSteps if self.maxSteps is not None else '%d seconds' % self.duration} ...")

        r = random.Random(self.randomSeed)
        chains = []
        costProgressions = []
        for i, schedule in enumerate(self._createSchedules(), start=1):
            self.log.info(f"Chain {i} uses {schedule}")
            chains.append(SAChain(stateFactory, schedule, opsAndWeights=self.opsAndWeights, randomSeed=r.randint(0, 1000)))
            costProgressions.append([])

        startTime = time.time()
        step = 0
        numChainSwaps = 0
        while True:
            timeElapsed = time.time() - startTime
            if (self.maxSteps is not None and step > self.maxSteps) or (self.duration is not None and timeElapsed > self.duration):
                break

            # take one step in each chain
            degreeOfCompletion = step / self.maxSteps if self.maxSteps is not None else timeElapsed / self.duration
            for chain in chains:
                chain.step(degreeOfCompletion)

            # check if neighbouring chains can be "swapped": if a high-temperature chain has a better state
            # than a low-temperature chain, swap them (by exchanging their schedules and swapping them
            # in the chains array, which shall always be in descending order of temperature)
            for idxHighTempChain in range(0, self.numChains-1):
                idxLowTempChain = idxHighTempChain+1
                highTempChain = chains[idxHighTempChain]
                lowTempChain = chains[idxLowTempChain]
                if highTempChain.state.cost.value() < lowTempChain.state.cost.value():
                    highTempChain.schedule, lowTempChain.schedule = lowTempChain.schedule, highTempChain.schedule
                    chains[idxLowTempChain] = highTempChain
                    chains[idxHighTempChain] = lowTempChain
                    numChainSwaps += 1

            if self.logCostProgression:
                for idxChain, chain in enumerate(chains):
                    costProgressions[idxChain].append(chain.state.cost.value())

            step += 1

        self.log.info(f"Number of chain swaps: {numChainSwaps}")
        if self.logCostProgression: self._costProgressions = costProgressions

        # apply best solution
        bestChainIdx = int(np.argmin([chain.bestCost.value() for chain in chains]))
        chains[bestChainIdx].applyBestState()

    def plotCostProgression(self):
        if not self.logCostProgression or self._costProgressions is None:
            raise Exception("No cost progression was logged")
        series = {}
        for scheduleParamStr, costProgression in zip(self._scheduleParamStrings, self._costProgressions):
            series[scheduleParamStr] = costProgression
        plt.figure()
        pd.DataFrame(series).plot()



import logging
import os
from typing import Sequence, Optional, Tuple

import matplotlib.figure
from matplotlib import pyplot as plt
import pandas as pd


_log = logging.getLogger(__name__)


class ResultWriter:
    _log = _log.getChild(__qualname__)

    def __init__(self, resultDir, filenamePrefix=""):
        self.resultDir = resultDir
        os.makedirs(resultDir, exist_ok=True)
        self.filenamePrefix = filenamePrefix

    def childWithAddedPrefix(self, prefix: str) -> "ResultWriter":
        """
        Creates a derived result writer with an added prefix, i.e. the given prefix is appended to this
        result writer's prefix

        :param prefix: the prefix to append
        :return: a new writer instance
        """
        return ResultWriter(self.resultDir, filenamePrefix=self.filenamePrefix + prefix)

    def childForSubdirectory(self, dirName: str):
        resultDir = os.path.join(self.resultDir, dirName)
        os.makedirs(resultDir, exist_ok=True)
        return ResultWriter(resultDir, filenamePrefix=self.filenamePrefix)

    def path(self, filenameSuffix: str, extensionToAdd=None, validOtherExtensions: Optional[Sequence[str]] = None):
        """
        :param filenameSuffix: the suffix to add (which may or may not already include the file extension "txt", which
            will be added if it is not already present)
        :param extensionToAdd: if not None, the file extension to add (without the leading ".") unless
            the extension to add or one of the extenions in validExtensions is already present
        :param validOtherExtensions: a sequence of valid other extensions (without the "."), only
            relevant if extensionToAdd is specified
        :return: the full path
        """
        if extensionToAdd is not None:
            addExt = True
            validExtensions = set(validOtherExtensions) if validOtherExtensions is not None else set()
            validExtensions.add(extensionToAdd)
            if validExtensions is not None:
                for ext in validExtensions:
                    if filenameSuffix.endswith("." + ext):
                        addExt = False
                        break
            if addExt:
                filenameSuffix += "." + extensionToAdd
        path = os.path.join(self.resultDir, f"{self.filenamePrefix}{filenameSuffix}")
        return path

    def writeTextFile(self, filenameSuffix, content):
        p = self.path(filenameSuffix, extensionToAdd="txt")
        self._log.info(f"Saving text file {p}")
        with open(p, "w") as f:
            f.write(content)
        return p

    def writeDataFrameTextFile(self, filenameSuffix, df: pd.DataFrame):
        p = self.path(filenameSuffix, extensionToAdd="df.txt", validOtherExtensions="txt")
        self._log.info(f"Saving data frame text file {p}")
        with open(p, "w") as f:
            f.write(df.to_string())
        return p

    def writeFigure(self, filenameSuffix, fig, closeFigure=False):
        """
        :param filenameSuffix: the filename suffix, which may or may not include a file extension, valid extensions being {"png", "jpg"}
        :param fig: the figure to save
        :param closeFigure: whether to close the figure after having saved it
        :return: the path to the file that was written
        """
        p = self.path(filenameSuffix, extensionToAdd="png", validOtherExtensions=("jpg",))
        self._log.info(f"Saving figure {p}")
        fig.savefig(p)
        if closeFigure:
            plt.close(fig)
        return p

    def writeFigures(self, figures: Sequence[Tuple[str, matplotlib.figure.Figure]], closeFigures=False):
        for name, fig in figures:
            self.writeFigure(name, fig, closeFigure=closeFigures)


import enum
import logging
import pickle
import threading
import time

import MySQLdb

from .cache import PersistentKeyValueCache


_log = logging.getLogger(__name__)


class MySQLPersistentKeyValueCache(PersistentKeyValueCache):
    class ValueType(enum.Enum):
        DOUBLE = ("DOUBLE", False)  # (SQL data type, isCachedValuePickled)
        BLOB = ("BLOB", True)

    def __init__(self, host, db, user, pw, valueType: ValueType, tableName="cache", deferredCommitDelaySecs=1.0):
        self.conn = MySQLdb.connect(host, db, user, pw)
        self.tableName = tableName
        self.maxKeyLength = 255
        self.deferredCommitDelaySecs = deferredCommitDelaySecs
        self._commitThread = None
        self._commitThreadSemaphore = threading.Semaphore()
        self._numEntriesToBeCommitted = 0

        cacheValueSqlType, self.isCacheValuePickled = valueType.value

        cursor = self.conn.cursor()
        cursor.execute(f"SHOW TABLES;")
        if tableName not in [r[0] for r in cursor.fetchall()]:
            cursor.execute(f"CREATE TABLE {tableName} (cache_key VARCHAR({self.maxKeyLength}) PRIMARY KEY, cache_value {cacheValueSqlType});")
        cursor.close()

    def set(self, key, value):
        key = str(key)
        if len(key) > self.maxKeyLength:
            raise ValueError(f"Key too long, maximal key length is {self.maxKeyLength}")
        cursor = self.conn.cursor()
        cursor.execute(f"SELECT COUNT(*) FROM {self.tableName} WHERE cache_key=%s", (key, ))
        storedValue = pickle.dumps(value) if self.isCacheValuePickled else value
        if cursor.fetchone()[0] == 0:
            cursor.execute(f"INSERT INTO {self.tableName} (cache_key, cache_value) VALUES (%s, %s)",
                (key, storedValue))
        else:
            cursor.execute(f"UPDATE {self.tableName} SET cache_value=%s WHERE cache_key=%s", (storedValue, key))
        self._numEntriesToBeCommitted += 1
        self._commitDeferred()
        #self.conn.commit()
        cursor.close()

    def get(self, key):
        cursor = self.conn.cursor()
        cursor.execute(f"SELECT cache_value FROM {self.tableName} WHERE cache_key=%s", (str(key), ))
        row = cursor.fetchone()
        if row is None:
            return None
        storedValue = row[0]
        value = pickle.loads(storedValue) if self.isCacheValuePickled else storedValue
        return value

    def _commitDeferred(self):
        self._lastUpdateTime = time.time()

        def doCommit():
            while True:
                time.sleep(self.deferredCommitDelaySecs)
                timePassedSinceLastUpdate = time.time() - self._lastUpdateTime
                if timePassedSinceLastUpdate >= self.deferredCommitDelaySecs:
                    _log.info(f"Committing {self._numEntriesToBeCommitted} cache entries to the database")
                    self.conn.commit()
                    self._numEntriesToBeCommitted = 0
                    return

        if self._commitThread is None or not self._commitThread.is_alive():
            self._commitThreadSemaphore.acquire()
            if self._commitThread is None or not self._commitThread.is_alive():
                self._commitThread = threading.Thread(target=doCommit, daemon=False)
                self._commitThread.start()
            self._commitThreadSemaphore.release()



import logging
import time
from typing import Callable

log = logging.getLogger(__name__)


def timed(method: Callable):
    """
    Decorator for execution timing
    """
    def timedExecution(*args, **kwargs):
        start = time.time()
        result = method(*args, **kwargs)
        end = time.time()
        log.info(f"Finished execution of {method.__name__} in {end-start:.2f}s")
        return result
    return timedExecution



from typing import Union, List, Dict, Any, Sequence
import re


def dictString(d):
    return ', '.join([f'{k}={v}' for k, v in d.items()])


def objectRepr(obj, memberNamesOrDict: Union[List[str], Dict[str, Any]]):
    def toString(x):
        if type(x) == dict:
            return "{" + dictString(x) + "}"
        else:
            return str(x)

    if type(memberNamesOrDict) == dict:
        membersDict = memberNamesOrDict
    else:
        membersDict = {m: toString(getattr(obj, m)) for m in memberNamesOrDict}
    return f"{obj.__class__.__name__}[{dictString(membersDict)}]"


def orRegexGroup(allowedNames: Sequence[str]):
    """

    :param allowedNames: strings to include as literals in the regex
    :return: raw string of the type (<name1>| ...|<nameN>), where special characters in the names have been escaped
    """
    allowedNames = [re.escape(name) for name in allowedNames]
    return r"(%s)" % "|".join(allowedNames)



from abc import ABC, abstractmethod
from typing import Tuple, Sequence, Generator, Optional, Union

from torch.autograd import Variable
import pandas as pd
import numpy as np
import torch

from .. import normalisation


def toTensor(d: Union[torch.Tensor, np.ndarray, list], cuda=False):
    if not isinstance(d, torch.Tensor):
        if isinstance(d, np.ndarray):
            d = torch.from_numpy(d)
        elif isinstance(d, list):
            d = torch.from_numpy(np.array(d))
        else:
            raise ValueError()
    if cuda:
        d.cuda()
    return d


class TensorScaler(ABC):
    @abstractmethod
    def cuda(self):
        """
        Makes this scaler's components use CUDA
        """
        pass

    @abstractmethod
    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Applies scaling/normalisation to the given tensor
        :param tensor: the tensor to scale/normalise
        :return: the scaled/normalised tensor
        """
        pass

    @abstractmethod
    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Applies the inverse of method normalise to the given tensor
        :param tensor: the tensor to denormalise
        :return: the denormalised tensor
        """
        pass


class TensorScalerFromVectorDataScaler(TensorScaler):
    def __init__(self, vectorDataScaler: normalisation.VectorDataScaler, cuda: bool):
        self.scale = vectorDataScaler.scale
        if self.scale is not None:
            self.scale = torch.from_numpy(vectorDataScaler.scale).float()
        self.translate = vectorDataScaler.translate
        if self.translate is not None:
            self.translate = torch.from_numpy(vectorDataScaler.translate).float()
        if cuda:
            self.cuda()

    def cuda(self):
        if self.scale is not None:
            self.scale = self.scale.cuda()
        if self.translate is not None:
            self.translate = self.translate.cuda()

    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.translate is not None:
            tensor -= self.translate
        if self.scale is not None:
            tensor /= self.scale
        return tensor

    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.scale is not None:
            tensor *= self.scale
        if self.translate is not None:
            tensor += self.translate
        return tensor


class TensorScalerIdentity(TensorScaler):
    def cuda(self):
        pass

    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor


class DataUtil(ABC):
    """Interface for DataUtil classes, which are used to process data for neural networks"""

    @abstractmethod
    def splitInputOutputPairs(self, fractionalSizeOfFirstSet) -> Tuple[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        """
        Splits the data set

        :param fractionalSizeOfFirstSet: the desired fractional size in
        :return: a tuple (A, B) where A and B are tuples (in, out) with input and output data
        """
        pass

    @abstractmethod
    def getOutputTensorScaler(self) -> TensorScaler:
        """
        Gets the scaler with which to scale model outputs

        :return: the scaler
        """
        pass

    @abstractmethod
    def getInputTensorScaler(self) -> TensorScaler:
        """
        Gets the scaler with which to scale model inputs

        :return: the scaler
        """
        pass

    @abstractmethod
    def modelOutputDim(self) -> int:
        """
        :return: the dimensionality that is to be output by the model to be trained
        """
        pass

    @abstractmethod
    def inputDim(self):
        """
        :return: the dimensionality of the input or None if it is variable
        """
        pass


class VectorDataUtil(DataUtil):
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame, cuda: bool, normalisationMode=normalisation.NormalisationMode.MAX_BY_COLUMN,
            differingOutputNormalisationMode=None):
        """
        :param inputs: the inputs
        :param outputs: the outputs
        :param cuda: whether to apply CUDA
        :param normalisationMode: the normalisation mode to use for inputs and (unless differingOutputNormalisationMode is specified) outputs
        :param differingOutputNormalisationMode: the normalisation mode to apply to outputs
        """
        if inputs.shape[0] != outputs.shape[0]:
            raise ValueError("Output length must be equal to input length")
        self.inputs = inputs
        self.outputs = outputs
        self.normalisationMode = normalisationMode
        self.inputValues = inputs.values
        self.outputValues = outputs.values
        inputScaler = normalisation.VectorDataScaler(self.inputs, self.normalisationMode)
        self.inputValues = inputScaler.getNormalisedArray(self.inputs)
        self.inputTensorScaler = TensorScalerFromVectorDataScaler(inputScaler, cuda)
        outputScaler = normalisation.VectorDataScaler(self.outputs, self.normalisationMode if differingOutputNormalisationMode is None else differingOutputNormalisationMode)
        self.outputValues = outputScaler.getNormalisedArray(self.outputs)
        self.outputTensorScaler = TensorScalerFromVectorDataScaler(outputScaler, cuda)

    def getOutputTensorScaler(self):
        return self.outputTensorScaler

    def getInputTensorScaler(self):
        return self.inputTensorScaler

    def splitInputOutputPairs(self, fractionalSizeOfFirstSet):
        n = self.inputs.shape[0]
        sizeA = int(n * fractionalSizeOfFirstSet)
        indices = list(range(n))
        indices_A = indices[:sizeA]
        indices_B = indices[sizeA:]
        A = self._inputOutputPairs(indices_A)
        B = self._inputOutputPairs(indices_B)
        return A, B

    def _inputOutputPairs(self, indices):
        n = len(indices)
        X = torch.zeros((n, self.inputDim()))
        Y = torch.zeros((n, self.outputDim()), dtype=self._torchOutputDtype())

        for i, outputIdx in enumerate(indices):
            inputData, outputData = self._inputOutputPair(outputIdx)
            if i == 0:
                if inputData.size() != X[i].size():
                    raise Exception(f"Unexpected input size: expected {X[i].size()}, got {inputData.size()}")
                if outputData.size() != Y[i].size():
                    raise Exception(f"Unexpected output size: expected {Y[i].size()}, got {outputData.size()}")
            X[i] = inputData
            Y[i] = outputData

        return X, Y

    def _inputOutputPair(self, idx):
        outputData = torch.from_numpy(self.outputValues[idx, :])
        inputData = torch.from_numpy(self.inputValues[idx, :])
        return inputData, outputData

    def inputDim(self):
        return self.inputs.shape[1]

    def outputDim(self):
        """
        :return: the dimensionality of the outputs (ground truth values)
        """
        return self.outputs.shape[1]

    def modelOutputDim(self):
        return self.outputDim()

    def _torchOutputDtype(self):
        return None  # use default (some float)


class ClassificationVectorDataUtil(VectorDataUtil):
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame, cuda, numClasses, normalisationMode=normalisation.NormalisationMode.MAX_BY_COLUMN):
        if len(outputs.columns) != 1:
            raise Exception(f"Exactly one output dimension (the class index) is required, got {len(outputs.columns)}")
        super().__init__(inputs, outputs, cuda, normalisationMode=normalisationMode, differingOutputNormalisationMode=normalisation.NormalisationMode.NONE)
        self.numClasses = numClasses

    def modelOutputDim(self):
        return self.numClasses

    def _torchOutputDtype(self):
        return torch.long

    def _inputOutputPairs(self, indices):
        # classifications requires that the second (1-element) dimension be dropped
        inputs, outputs = super()._inputOutputPairs(indices)
        return inputs, outputs.view(outputs.shape[0])


class TorchDataSet:
    @abstractmethod
    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False) -> Generator[Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor], None, None]:
        pass

    @abstractmethod
    def size(self) -> Optional[int]:
        pass


class TorchDataSetProvider:
    def __init__(self, inputTensorScaler: Optional[TensorScaler] = None, outputTensorScaler: Optional[TensorScaler] = None,
            inputDim: Optional[int] = None, modelOutputDim: int = None):
        if inputTensorScaler is None:
            inputTensorScaler = TensorScalerIdentity()
        if outputTensorScaler is None:
            outputTensorScaler = TensorScalerIdentity()
        if modelOutputDim is None:
            raise ValueError("The model output dimension must be provided")
        self.inputTensorScaler = inputTensorScaler
        self.outputTensorScaler = outputTensorScaler
        self.inputDim = inputDim
        self.modelOutputDim = modelOutputDim

    @abstractmethod
    def provideSplit(self, fractionalSizeOfFirstSet: float) -> Tuple[TorchDataSet, TorchDataSet]:
        pass

    def getOutputTensorScaler(self) -> TensorScaler:
        return self.outputTensorScaler

    def getInputTensorScaler(self) -> TensorScaler:
        return self.inputTensorScaler

    def getModelOutputDim(self) -> int:
        """
        :return: the number of output dimensions that would be required to be generated by the model to match this dataset.
        """
        return self.modelOutputDim

    def getInputDim(self) -> Optional[int]:
        """
        :return: the number of output dimensions that would be required to be generated by the model to match this dataset.
            For models that accept variable input sizes (such as RNNs), this may be None.
        """
        return self.inputDim


class TorchDataSetFromTensors(TorchDataSet):
    def __init__(self, x: torch.Tensor, y: Optional[torch.Tensor], cuda: bool):
        if y is not None and x.shape[0] != y.shape[0]:
            raise ValueError("Tensors are not of the same length")
        self.x = x
        self.y = y
        self.cuda = cuda

    def iterBatches(self, batchSize: int, shuffle: bool = False, inputOnly=False) -> Generator[Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor], None, None]:
        tensors = (self.x, self.y) if not inputOnly and self.y is not None else (self.x,)
        yield from self._get_batches(tensors, batchSize, shuffle)

    def _get_batches(self, tensors: Sequence[torch.Tensor], batch_size, shuffle):
        length = len(tensors[0])
        if shuffle:
            index = torch.randperm(length)
        else:
            index = torch.LongTensor(range(length))
        start_idx = 0
        while start_idx < length:
            end_idx = min(length, start_idx + batch_size)
            excerpt = index[start_idx:end_idx]
            batch = []
            for tensor in tensors:
                if len(tensor) != length:
                    raise Exception("Passed tensors of differing lengths")
                t = tensor[excerpt]
                if self.cuda:
                    t = t.cuda()
                batch.append(Variable(t))
            if len(batch) == 1:
                yield batch[0]
            else:
                yield tuple(batch)
            start_idx += batch_size

    def size(self):
        return self.y.shape[0]


class TorchDataSetProviderFromDataUtil(TorchDataSetProvider):
    def __init__(self, dataUtil: DataUtil, cuda: bool):
        super().__init__(inputTensorScaler=dataUtil.getInputTensorScaler(), outputTensorScaler=dataUtil.getOutputTensorScaler(),
            inputDim=dataUtil.inputDim(), modelOutputDim=dataUtil.modelOutputDim())
        self.dataUtil = dataUtil
        self.cuda = cuda

    def provideSplit(self, fractionalSizeOfFirstSet: float) -> Tuple[TorchDataSet, TorchDataSet]:
        (x1, y1), (x2, y2) = self.dataUtil.splitInputOutputPairs(fractionalSizeOfFirstSet)
        return TorchDataSetFromTensors(x1, y1, self.cuda), TorchDataSetFromTensors(x2, y2, self.cuda)



import copy
import logging
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from sklearn import compose

from ..vector_model import VectorRegressionModel, VectorClassificationModel

_log = logging.getLogger(__name__)


def createSkLearnModel(modelConstructor, modelArgs, outputTransformer=None):
    model = modelConstructor(**modelArgs)
    if outputTransformer is not None:
        model = compose.TransformedTargetRegressor(regressor=model, transformer=outputTransformer)
    return model


class AbstractSkLearnVectorRegressionModel(VectorRegressionModel, ABC):
    """
    Base class for models built upon scikit-learn's model implementations
    """
    _log = _log.getChild(__qualname__)

    def __init__(self, modelConstructor, **modelArgs):
        """
        :param modelConstructor: the sklearn model constructor
        :param modelArgs: arguments to be passed to the sklearn model constructor
        """
        super().__init__()
        self.sklearnInputTransformer = None
        self.sklearnOutputTransformer = None
        self.modelConstructor = modelConstructor
        self.modelArgs = modelArgs

    def withSkLearnInputTransformer(self, sklearnInputTransformer) -> __qualname__:
        """
        :param sklearnInputTransformer: an optional sklearn preprocessor for normalising/scaling inputs
        :return: self
        """
        self.sklearnInputTransformer = sklearnInputTransformer
        return self

    def withSkLearnOutputTransformer(self, sklearnOutputTransformer):
        """
        :param sklearnOutputTransformer: an optional sklearn preprocessor for normalising/scaling outputs
        :return: self
        """
        self.sklearnOutputTransformer = sklearnOutputTransformer
        return self

    def _transformInput(self, inputs: pd.DataFrame, fit=False) -> pd.DataFrame:
        if self.sklearnInputTransformer is None:
            return inputs
        else:
            inputValues = inputs.values
            shapeBefore = inputValues.shape
            if fit:
                inputValues = self.sklearnInputTransformer.fit_transform(inputValues)
            else:
                inputValues = self.sklearnInputTransformer.transform(inputValues)
            if inputValues.shape != shapeBefore:
                raise Exception("sklearnInputTransformer changed the shape of the input, which is unsupported. Consider using an a DFTSkLearnTransformer in inputTransformers instead.")
            return pd.DataFrame(inputValues, index=inputs.index, columns=inputs.columns)

    def _updateModelArgs(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to modelArgs

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _fit(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        inputs = self._transformInput(inputs, fit=True)
        self._updateModelArgs(inputs, outputs)
        self._fitSkLearn(inputs, outputs)

    @abstractmethod
    def _fitSkLearn(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        pass

    def _predict(self, x: pd.DataFrame):
        inputs = self._transformInput(x)
        return self._predictSkLearn(inputs)

    @abstractmethod
    def _predictSkLearn(self, inputs: pd.DataFrame):
        pass


class AbstractSkLearnMultipleOneDimVectorRegressionModel(AbstractSkLearnVectorRegressionModel, ABC):
    """
    Base class for models which use several sklearn models of the same type with a single
    output dimension to create a multi-dimensional model (for the case where there is more than one output dimension)
    """
    def __init__(self, modelConstructor, **modelArgs):
        super().__init__(modelConstructor, **modelArgs)
        self.models = {}

    def __str__(self):
        if len(self.models) > 0:
            modelStr = str(next(iter(self.models.values())))
        else:
            modelStr = f"{self.modelConstructor.__name__}{self.modelArgs}"
        return f"{self.__class__.__name__}[{modelStr}]"

    def _fitSkLearn(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        for predictedVarName in outputs.columns:
            _log.info(f"Fitting model for output variable '{predictedVarName}'")
            model = createSkLearnModel(self.modelConstructor,
                    self.modelArgs,
                    outputTransformer=copy.deepcopy(self.sklearnOutputTransformer))
            model.fit(inputs, outputs[predictedVarName])
            self.models[predictedVarName] = model

    def _predictSkLearn(self, inputs: pd.DataFrame) -> pd.DataFrame:
        results = {}
        for varName in self.models:
            results[varName] = self.models[varName].predict(inputs)
        return pd.DataFrame(results)


class AbstractSkLearnMultiDimVectorRegressionModel(AbstractSkLearnVectorRegressionModel, ABC):
    """
    Base class for models which use a single sklearn model with multiple output dimensions to create the multi-dimensional model
    """
    def __init__(self, modelConstructor, **modelArgs):
        super().__init__(modelConstructor, **modelArgs)
        self.model = None

    def __str__(self):
        if self.model is not None:
            modelStr = str(self.model)
        else:
            modelStr = f"{self.modelConstructor.__name__}{self.modelArgs}"
        return f"{self.__class__.__name__}[{modelStr}]"

    def _fitSkLearn(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        if len(outputs.columns) > 1:
            _log.info(f"Fitting a single multi-dimensional model for all {len(outputs.columns)} output dimensions")
        self.model = createSkLearnModel(self.modelConstructor, self.modelArgs, outputTransformer=self.sklearnOutputTransformer)
        outputValues = outputs.values
        if outputValues.shape[1] == 1:  # for 1D output, shape must be (numSamples,) rather than (numSamples, 1)
            outputValues = np.ravel(outputValues)
        self.model.fit(inputs, outputValues)

    def _predictSkLearn(self, inputs: pd.DataFrame) -> pd.DataFrame:
        Y = self.model.predict(inputs)
        return pd.DataFrame(Y, columns=self.getModelOutputVariableNames())


class AbstractSkLearnVectorClassificationModel(VectorClassificationModel, ABC):
    def __init__(self, modelConstructor, **modelArgs):
        """
        :param modelConstructor: the sklearn model constructor
        :param modelArgs: arguments to be passed to the sklearn model constructor
        """
        super().__init__()
        self.modelConstructor = modelConstructor
        self.sklearnInputTransformer = None
        self.sklearnOutputTransformer = None
        self.modelArgs = modelArgs
        self.model = None

    def withSkLearnInputTransformer(self, sklearnInputTransformer) -> __qualname__:
        """
        :param sklearnInputTransformer: an optional sklearn preprocessor for normalising/scaling inputs
        :return: self
        """
        self.sklearnInputTransformer = sklearnInputTransformer
        return self

    def withSkLearnOutputTransformer(self, sklearnOutputTransformer):
        """
        :param sklearnOutputTransformer: an optional sklearn preprocessor for normalising/scaling outputs
        :return: self
        """
        self.sklearnOutputTransformer = sklearnOutputTransformer
        return self

    def __str__(self):
        if self.model is None:
            strModel = f"{self.modelConstructor.__name__}{self.modelArgs}"
        else:
            strModel = str(self.model)
        return f"{self.__class__.__name__}[{strModel}]"

    def _updateModelArgs(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to modelArgs

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _fitClassifier(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        inputValues = self._transformInput(inputs, fit=True)
        self._updateModelArgs(inputs, outputs)
        self.model = createSkLearnModel(self.modelConstructor, self.modelArgs, self.sklearnOutputTransformer)
        _log.info(f"Fitting sklearn classifier of type {self.model.__class__.__name__}")
        self.model.fit(inputValues, np.ravel(outputs.values))

    def _transformInput(self, inputs: pd.DataFrame, fit=False) -> np.ndarray:
        inputValues = inputs.values
        if self.sklearnInputTransformer is not None:
            if fit:
                inputValues = self.sklearnInputTransformer.fit_transform(inputValues)
            else:
                inputValues = self.sklearnInputTransformer.transform(inputValues)
        return inputValues

    def _predict(self, x: pd.DataFrame):
        inputValues = self._transformInput(x)
        Y = self.model.predict(inputValues)
        return pd.DataFrame(Y, columns=self._predictedVariableNames)

    def _predictClassProbabilities(self, x: pd.DataFrame):
        inputValues = self._transformInput(x)
        Y = self.model.predict_proba(inputValues)
        return pd.DataFrame(Y, columns=self._labels)

    def get_params(self, deep=True):
        return self.model.get_params(deep=deep)

    def set_params(self, **params):
        self.model.set_params(**params)

